/*
 JUMA-PA100 Amplifier Controller
 Juha Niinikoski, OH2NLT 06.07.2008

 Microchip MPLAB + C30 compiler
 MCU dsPIC30F6014A

 Auto F-sense code 08.07.2008
 Boot loader & Temp scaling 09.07.2008
 EEPROM structures & Encoder simulator 17.07.2008
 Power meter fast attack, slow release 04.08.2008
 Improved F-sense code 05.08.2008
 Yaesu 817 band select tests, check_io_board_type() removed 07.08.2008
 TRX2 serial band set protocol 09.08.2008
 7,3728MHz XTAL, repeat for UP/DN buttons 13.08.2008
 SWR Alarm, Alarm etc, setup displays 13.08.2008
 PA100 - TRX2 serial protocol 14.08.2008
 Yaesu FT817 band voltage table matched with 4M7 input resistance 16.08.2008
 SWR alarm fine tuning 22.08.2008
 Display clean up, Auto/Man logic changed, Alarm beep added 12.11.2008 v0.04
 RS232 band set & query  active only if TRX2 protocol is selected 15.11.2008
 RS232 printouts only in the test mode, do not send garbage to the TRX2 15.11.2008
 Some text changed, default calibration values changed 15.11.2008
 TX capability added to the service mode 15.11.2008 v1.00 
 PA100 temp meter degree sign added to the soft fonts #4 16.11.2008 v1.00
 TRX2 protocol poll rate adjust added 17.11.2008 v1.00
 Alarms force normal display, some text changes 18.11.2008
 UART module data type cast corrections 18.11.2008
 Band +/- end of range tones changed, OPER ==> STBY while TX on corrected 19.11.2008 v1.01
 Band info, ! = out of band, ? = not known, TX/tx display added 22.11.2008	v1.02
 Poll timer changed from calval to defval, TRX2 poll response timeout, some display text changes 30.11.2008 v1.03
 Service mode factory setup function corrected 16.12.2008 v1.04
 ---------------------------------- 5B4AIY Modifications ----------------------------------
 Version 1.05h
 Various typographical errors fixed, minor display formatting improvements - 5B4AIY 23/AUG/2011
 Changed current measurement to a 50 sample running average - 5B4AIY 24/SEP/2012
 Added selectable graphical display of measurements in relation to their trip limits. - 5B4AIY 04/OCT/2012
 Cleaned up some code, removed redundant variables. - 5B4AIY 04/OCT/2012
 Build 6.
 Added feature to allow PA-100D to work with the Elecraft KX3 transceiver.
 Build 7.
 Removed redundant code related to fsense logic - Frequency Sense board is always present.
 Minor changes to prompt messages for Calibration and User Configuration mmenus.
 Build 8.
 Changed polling so that it now increments in 1 second intervals from 0 - 10 seconds.
 As before, 0 = Disabled. - 5B4AIY 15/OCT/2012
 Build 10.
 Revised EEPROM layout and changed id_mult, batt_mult, rev_pwr_mult and fwd_pwr_mult to integers. A.Ryan 5B4AIY 22/OCT/2012
 Build 14.
 Revised peak hold time from 350mS to 1S, and fast decay time from 50mS to 20mS.
 Optimised service and configuration routines. A.Ryan - 5B4AIY
 Build 15.
 Commented out all the loop counter code used to test the loop execution time. A.Ryan 23/NOV/2012
 Build 16.
 Replaced several one-line functions with equivalent macro definitions in juma-pa100.h A.Ryan - 5B4AIY - 27/NOV/2012
 Build 17.
 Replaced inelegant attenuator relay switch and fan speed switch code with more elegant version.
 34 lines of switch/case code replaced with 4 lines of conditional expression code.
 Replaced loop invariant address calculations with pointer set during boot initialisation. (rs232_mode)
 Re-organised the order of the frequency sense modes.
 Modified the way the serial test mode is selected. If either the TRX-2 or KX3 frequency sense modes are selected,
 then the serial test mode selection is skipped. If either the F-Sense or FT-817 frequency sense modes are selected,
 then the serial test page is displayed, and the test mode can be turned on or off. A.Ryan - 5B4AIY - 7/JAN/2013

 Version 1.05i
 Build 1
 Moved the Polling Interval time setting from the calibration menu to the configuration menu. With the
 option of using either the TRX-2 or the KX3 serial port auto-band detect modes, it made more sense to
 combine this with their polling interval settings rather than having to switch to the calibration setting.
 The logic now skips the polling interval setting page if either the Frequency Sense or the FT-817 auto band detect
 mode is selected, or skips the Serial Test mode setting page if either the TRX-2 or KX3 mode is selected.
 Modified the Service Mode Default so that it uses a prompted exit as with the normal save settings.
 Modified the power up boot sequence so that the long blank screen when selecting the Service Mode is now
 replaced with the normal sign-on splash screen. In addition, the save_settings() function has been modified
 to check for the BAND+ pushbutton pressed as well as the PWR button, which greatly simplified the default
 restore option when powering up. A.Ryan - 5B4AIY - 9/JAN/2013
 Build 2
 Moved all embedded literal prompts and messages to constant character strings.
 Modified the way the main display page is selected. If the KEY status is used directly rather than assigning
 its value to an internal variable, then this takes 54 bytes more code! A.Ryan - 5B4AIY - 18/JAN/2013
 Build 3
 Added a test in four places to inhibit printouts to the serial port if the Auto-Band Detect mode is set to
 either TRX-2 or KX3, or if the Serial Test Mode is set to OFF. This prevents spurious characters being
 transmitted to the TRX-2 or the KX3, which might confuse them. Modified the operation of the command time-out
 timer so that it is disabled if the polling timer is enabled. This permits the user to perform a test to
 verify that the KX3 command parser is working by sending the polling response string from a terminal.
 For example, set the Auto-Band Detect to KX3. Select the AUTO mode of band detection.
 Enable the Polling Timer and set it to 10 seconds. Connect the serial port to a PC running a terminal program.
 Verify that every 10 seconds you receive the polling request: IF; then enter the text: IF14000000; and the
 amplifier should immediately switch to 20m. Wait for the polling time-out to occur, and the band should change
 to: ? By triggering the TX data line by grounding the tip of a standard 3.5mm stereo patch cable connected to
 the T/R socket on the rear of the amplifier, you can also verify that the TX Enable/Disable logic is working.
 Select MANUAL, and OPER. Ground the tip. The amplifier should show TX. Now select AUTO, ground the tip, and
 the amplifier should show NOP. Wait for the poll request, and send: IF7000000; and the amplifier should switch
 to 40m and TX. When the polling request next times-out, the state should switch to NOP if the tip is still
 grounded.
 Removed set_cgram() from lcd-trx2.c - never referenced. Optimised set_ch_bits() A.Ryan - 5B4AIY - 22/JAN/2013
 Build 4
 Restructured both the clr_rx_buffer() and clear_buffer() functions to use a simple DO-WHILE loop rather than
 the more usual FOR-NEXT loop. This used 15 bytes less memory in each case. A.Ryan - 5B4AIY - 25/JAN/2013
 Build 5
 Removed send_band_data() from serial_pa100.c Only referenced once, replaced with single line of code.
 Removed ms_counter from timers_pwm.c, never used. Changed maximum backlight level from 1100 to 1000,
 there was very little change above 1000.
 Set initial backlight and contrast to their default values. A.Ryan - 5B4AIY - 04/MAR/2013
 Build 6
 Added function display_screen() to bring together 5 sections of identical code, as well as improve
 the usage of the data section by removing a duplicate prompt. Re-wrote the section of code that
 displays the STDBY/TX/NOP/OPER message. Replaced SWITCH/CASE construct with simpler IF/ELSE. A.Ryan - 04/JUN/2013

 Version 1.05j
 Build 1
 Re-wrote set_relay() function to eliminate large amounts of duplicated code. Added a feature to the serial test
 suite to display the currently selected output filter. A.Ryan - 5B4AIY - 07/JUN/2013
 Build 2
 Further simplification of the set_relay() function, reducing it to only 8 lines of code, as compared
 to 57 in the original. A.Ryan - 5B4AIY - 08/JUN/2013
 Build 3
 Changed ammeter to a peak reading long hold meter. The graphic meter is still an instantaneous reading
 meter. The peak reading ammeter is disabled in the Service Mode. A.Ryan - 5B4AIY - 11/JUN/2013

 Version 1.05k
 Build 1
 With the availability of an actual KX3, it was possible to optimise and improve the compatibility. With
 the previous firmware when using the KX3's AUTO-INFORM mode set to ANT-CTRL, the band change response time
 of the PA-100D was often as long as 5 seconds. This was caused by the KX3 sending the IF data packet
 last, and this could take 5 seconds. An examination of the KX3's response in this mode showed that the
 first data packet was the FA VFO-A frequency data, followed by the FB VFO-B data. It thus made sense to
 recognise these packets and set the PA-100D's band accordingly. This version of the firmware modifies the
 operation. If polling is enabled, then the only acceptable response string is the IF data packet, and this
 must be received within the prescribed time-out period to be valid. The time-out period is the polling interval.
 If polling is disabled, then the KX3 must be set to AUTO-INFORM with the ANT-CTRL mode selected. In this mode
 the acceptable responses for a band change is the FA/FB VFO frequency data, as well as the IF information packet.
 In the AUTO-INFORM mode, a message timer of 500mS is started when the first valid character is received.
 This is more than sufficient for the longest message at the slowest speed. It simply ensures that
 if serial communications is interupted then the PA-100D does not simply wait forever for a response but will
 clear the buffer and restart the serial command handler.
 An example of a KX3 Auto-Inform data packet is:
 Select 20m:
 FA00014060000;
 FB00014060400;
 RT0;
 RO-0000;
 MD2;
 MD$2;
 IF00014060000     -000000 0002000001 ;
 A. Ryan - 5B4AIY - 14/JUL/2013
 Build 2
 After examining the KX3 data packets, it was obvious that only the FA packet needs to be examined. Whenever
 a band change occurs and AUTO-INFORM is on, then the FA packet is always sent first. There is no need to
 examine any other data packet. This permitted a considerable simplification of the KX3 parser, eliminating
 the check_buffer() function, and allowing the parser to automatically select the FA data and reject everything
 else. I also added a 'hidden' debugging function. If the BAND- button is pressed, and then the power turned on,
 the KX3 message timeout value is increased from 500mS to 5000mS so that the PA-100D can be tested with a
 terminal program when in the KX3 mode. To ensure that when polling is disabled the PA-100D still selects a
 valid band, the first time the serial port function is exercised a forced FA query command is transmitted,
 and the KX3 will respond with the current setting. This will cause the band to be set correctly. To ensure
 that this happens, power up the KX3 first and let it initialise, then power up the PA-100D. A.Ryan 14/JUL/2013
 Build 3
 Changed the Yes_No[] prompt message. Since the PWR button is used for the 'No' response, and it is located on the
 left-hand side of the amplifier's control panel, and the BAND+ button for the 'Yes' response, located on the
 right-hand side of the control panel, it made more sense for the message to be so arranged. In addition,
 the BAND- button response has been eliminated. A.Ryan - 5B4AIY - 31/AUG/2013
 Build 4
 Modified the set_value() function to avoid duplicative address calculations. The second argument is now a pointer
 to an integer rather than the value itself, and the function now modifies the value directly via this pointer, thus
 avoiding needless address duplication. This saved 45 bytes. Removed the null() function. This simply discarded any characters
 in the USART1 buffer if the test mode was off, but the same effect can be obtained by simply resetting the buffer pointers,
 thus eliminating a redundant function, which saved an additional 21 bytes. A.Ryan - 5B4AIY - 21/NOV/2013

 1.05m
 NOTE: There is no version l.05l - this is because the l and the figure 1 are not easily distinguishable.

 Build 1
 Rationalised the way the User Configuration and System Calibration Settings are saved. Previously the User Configuration
 settings were saved by pressing and holding the DISPLAY/CONFIG button, and the System Calibration Settings by pressing
 the OPER button. This version rationalises this by using the OPER button for both. A.Ryan - 5B4AIY - 03/DEC/2013
 Build 2
 Added Auto-Increment to the User Configuration and System Calibration Menus. To use this feature, press and hold the
 DISPLAY/CONFIG button. After a short delay, the pages will automatically increment. A.Ryan - 5B4AIY - 04/DEC/2013
 Build 3
 Modified serial_kx3() to eliminate a switch-case, replaced with two IF-THEN constructs. A.Ryan - 5B4AIY - 04/DEC/2013
 Build 4
 Optimised increment_cal_page() and increment_cfg_page() by combining their logic tests into a single expression.
 This saved 18 bytes. A.Ryan - 5B4AIY - 06/DEC/2013
 Build 5
 Removed header file: lcd-trx2.h and incorporated its contents into: juma-pa100.h. Fixed low-level re-definitrion bug.
 In file: juma-pa100.c the symbol S_BAR was re-defined, although it was never used in this module. A.Ryan - 5B4AIY - 27/DEC/2013
 Build 6
 Removed set_cur_lcd() and replaced with lcd_cmd(), changing all the cursor and line positioning commands to their
 instrinsic values. Optimised initlcd(). A.Ryan - 5B4AIY - 02/JAN/2014
 Build 7
 Relocated local static variable old_data to exchg_data_spi1(). Removed local variables spi_rx_data, never used.
 Optimised SPI1 Interrupt, added explanation of how the SPI bus operates to spi1.c Removed SPI code - not used. A.Ryan - 5B4AIY - 3/JAN/2014
 Build 8
 Simplified the code to show band information. A.Ryan - 5B4AIY - 26/JAN/2014
 Build 9
 Re-wrote disp_meter() considerably simplifying it. Rationalised the SWR display. A.Ryan - 5B4AIY - 27/JAN/2014
 Build 10
 Changed the PAGE_CHANGE constant to 1000mS. A.Ryan - 5B4AIY - 27/JAN/2014
 Build 11
 Fixed calibration display bug. Voltage was displayed in mV rather than Volts. A.Ryan - 5B4AIY - 31/JAN/2014
 Build 12
 Replaced EEPROM variable: eeprom.defval.pa_state with global flag: pa_state. Added new feature, Start-Up Page Select. This utilises
 the variable: eeprom.defval.pa_state. It is defined as: start_page as a macro in juma-pa100.h A.Ryan - 5B4AIY - 07/FEB/2014
 Build 13
 Changed the LONG_PUSH to 650mS, and the VERY_LONG_PUSH to 1,500mS. A.Ryan - 5B4AIY - 08/MAR/2014
 Build 14
 Used svc_flag to both remain in and exit from service(). A.Ryan - 5B4AIY - 22/MAR/2014
 Build 15
 Modified putch() in uart.c to eliminate redundant mask operation. Optimised sound test in serial_test.c. Fixed minor bug. If a
 loopback plug is inserted into the RS-232 port and the RS-232 loopback test is selected, it would continue to run forever. This
 has now been fixed. A.Ryan - 5B4AIY - 07/APR/2014
 Build 16
 Eliminated redundant Get_Character() function in Serial Test Suite, and added a buffer clear to the illegal character
 code. Eliminated redundant variable ultemp. A.Ryan - 5B4AIY - 17/APR/2014

 1.05p
 Build 1
 Complete re-write of button recognition and function dispatch logic. Removed redundant logic tests, superfluous variables
 and function calls. Although this did not have as dramatic an effect as with the TRX-2, it was still a worthwhile exercise
 and resulted in more compact code, and more elegant construction. Modified the auto-increment feature to give a longer beep
 when switching back to the first page, as well as adding an extra delay. The same sounds and logic is used for the
 Calibration & Service and the User Configuration functions. Added code to ignore DISPLAY/OPER/AUTO/BAND-/BAND+ buttons in
 alarm state. A.Ryan - 5B4AIY - 28/APR/2014
 Build 2
 Added code to ignore UP/DOWN buttons in alarm state. Only the PWR button is now recognised. Modified the 1mS interrupt
 handler to use a DO-WHILE loop in the band select module with a pre-increment loop counter, as this uses fewer machine cycles.
 Similarly, the functions eval_band() and reset_fsense() were also modified. A.Ryan - 5B4AIY - 29/APR/2014
 Build 3
 One of the persistent annoyances is the relay chatter when using the frequency sense mode and SSB. An examination of the
 fsense test output reveals why. During speech you can sometimes get significant counts in the lower frequency bins because
 of the syllabic rate of speech. This is particularly evident when using only a low power to drive the amplifier, and it
 causes a low-frequency band filter to be temporarily selected, because the count in the correct band is zero, whereas
 there is a count in a lower band because of this mis-counting occuring during speech syllables. To effect a cure, the
 eval_band() function was greatly modified. A new static variable, last_band, has been introduced, and during the
 transmit mode this is set to the measured frequency band. If on subsequent iterations a higher frequency band is detected,
 then this is updated, as is eeprom.defval.band. In this way the band is always set to the highest frequency and never
 reset to a lower band. In the receive mode, last_band is reset for the next transmit cycle. The frequency sense auto band
 detect works in STANDBY as well as OPERATE, whether there is an external key signal, or whether using the JUMA transmit
 key signal. A.Ryan - 5B4AIY - 29/APR/2014
 Build 4
 Inhibited manual band changes in transmit. A.Ryan - 5B4AIY - 01/MAY/2014
 Build 5
 Re-wrote display_pwr_meter() and re-named it scaled_value(). It was only called once and the SWITCH-CASE construct could
 be replaced with a simple function call. Optimised getch() and kbhit(). Minor improvements to the LCD Write ASCII and
 Write HEX tests. A.Ryan - 5B4AIY - 23/MAY/2014
 Build 6
 Re-wrote reset_fsense() to eliminate redundant flag. The variable freq_sample_ctr already contains the required information
 to determine when a valid number of counts has been accumulated. Modified the routine in the 1mS timer interrupt to suit.
 Modified the interrupt handler so that if a frequency is detected that is higher than 30MHz, the amplifier is immediately
 forced to the STANDBY state, and the band set to NOT_KNOWN. A.Ryan - 5B4AIY - 11/JUN/2013
 Build 7
 Optimised clear_buffer() in serial_pa100.c and simplified serial_pa100(). A.Ryan - 5B4AIY - 13/JUN/2014
 Build 8
 Combined the KX3 and PA-100 serial buffers. Removed the redundant KX3 clear buffer function. A.Ryan - 5B4AIY - 06/AUG/2014
 Build 9
 Modified wait_lcd_ready() in lcd-trx2.c to remove redundant assignment, and use simpler DO-WHILE loop. Shortened the Read Data
 setup delay from 2uS to 1uS in wait_lcd_ready(). The maximum delay to data valid is 320nS, the minimum available delay time
 is 1uS. Removed redundant Restore Defaults page in System Calibration & Setup. This facility is already available by using
 the MODE button from the initial power-on. A.Ryan - 5B4AIY - 17/AUG/2014

 1.05q
 Build 1
 Changed the way the User Configuration menu is initially displayed. In the previous versions, when the User Configuration menu
 was selected, the current configuration page was displayed. In this version, the message: 'User Configuration' is displayed as
 long as the DISPLAY/CONFIG button is depressed. When it is released, the current configuration page is displayed. A subsequent
 long push will auto-increment the configuration pages, as before. A.Ryan - 5B4AIY - 14/SEP/2014
 Build 2
 Fixed bug in serial test suite. If using the Yaesu band information, there would be an Address Trap error when the band value
 was greater than 7 - insufficient entries in the table - fixed. A.Ryan - 5B4AIY - 08/OCT/2014

 1.05r
 Build 1
 Eliminated an unused prompt in service.c, and simplified the TX switch logic. Forced the use of the Frequency Sense mode,
 and the minimum gain for the ammeter and power meter calibration. Changed logic so that the amplifier will only switch to
 the transmit mode when either the ammeter or RF power calibration pages are selected. Forced Calibration & Setup abort if
 an alarm is detected. Saved the current band gain settings on entry, and restored them on exit to the Calibration & Setup
 mode. A.Ryan - 5B4AIY - 27/OCT/2014
 Build 2
 Modified disp_id() so that the current is always displayed, even in the receive mode. Since the current measured is the
 total amplifier current, there is no reason not to display it at all times if this display mode is selected. The display
 has been updated so that if there is a change of amplifier state then the display is immediately updated without waiting
 for the hold time to expire, and the hold counter is reset. A.Ryan - 5B4AIY - 28/OCT/2014
 Build 3
 In the previous firmware, when the Service Mode is selected, there is a low-level bug that somtimes causes a mode and/or
 a VFO change to the TRX-2. The conditions have to be very specific: the Juma TRX-2 has to have the TRX2 serial port
 protocol selected; the speed has to be 9600 Baud; the serial port cable has to be connected to the amplifier's RS-232
 port; the amplifier's User Configuration frequency select mode has to be either F-SENSE or FT-817; the Serial Test flag
 has to be ON; and the Juma TRX-2 mode has to be TUNE and VFO-A. If these conditions are met, then when exiting the Service
 Mode after a short delay of about 1 or 2 seconds, the TRX-2 mode will change to USB, and sometimes VFO-B will also be
 selected. However, if the amplifier's Serial Test flag is set to OFF, this does not occur. The reason for this is that
 when the transceiver's band is changed, or if it is keyed this will send a status update command to the amplifier, which
 will be saved in the serial buffer. When the Service Mode is exited, the amplifier's Serial Test Suite will retrieve
 these saved commands from the buffer and attempt to implement them. This will provoke a series of error messages, and
 some of these could be interpreted by the TRX-2 as mode or VFO change commens. Although the same effect could be achieved
 by simply clearing the USART1 queue, when exiting the Service Mode if the Serial Test Suite is enabled and the TRX-2
 is in the TRX2 protocol mode, then the continuing status updates would cause more spurious operation, so it is much
 better to simply set the flag OFF. The user can then reset it for when it is required. A.Ryan - 5B4AIY - 10/NOV/2014

 1.06a	This version uses additional EEPROM and will cause a checksum error on initial load.
 Build 1
 Added feature to allow display of RF O/P power in either Watts or dBm. A.Ryan - 5B4AIY - 13/NOV/2014
 Build 2
 Added feature to allow the use of the F-SENSE mode to display the input frequency. (See notes in timers_pwm.c for details
 of the changes and an explanation of the algorithm used.) A.Ryan - 5B4AIY - 15/NOV/2014
 Build 3
 Added calibration feature for the low power range of the RF power meter. This uses an additional word of EEPROM
 and so will cause a checksum error on initial load. Fixed minor bug in Serial Test suite. When using the F-SENSE
 automatic mode of band detection if the ADC test is invoked, the band resets to UNKNOWN at the conclusion of the
 test. This was because the FT-817 band evaluation function was invoked, and this would reset the current band
 based upon any voltage detected on this port. Modified eval_y817() and added function get_817() to
 fix this problem. Changed power meter display so that you no longer need to key the amplifier in order to display
 output power. This allows a transceiver's output power to be displayed in the STANDBY mode. Also modified the
 lower limit of the display to be +26dBm/400mW. Measurements indicate that the frequency counter will read reliably
 with input powers of as low as 100mW, and the low power reading is accurate down to 500mW. Equally, the counter
 will display frequencies up to 54MHz with only a slightly reduced input sensistivity. A.Ryan - 5B4AIY - 21/NOV/2014
 Build 4
 Changed test in disp_fwd_pwr() from: if(out_pwr > peak_pwr) to: if(out_pwr >= peak_pwr) to eliminate an annoying
 display blink when the power is constant. Improved the accuracy of the SWR indication by changing calculation
 from long integer to floating-point. A.Ryan - 5B4AIY - 23/NOV/2014
 Build 5
 Introduced a new global variable vswr. This is a double, and takes the value from calc_swr(). The value is converted
 to an int and assigned to swr. vswr is used in the call to disp_meter() to save having to convert from a double
 to an int and then back to a double, and so should marginally improve the accuracy of the meter. A.Ryan - 5B4AIY - 25/NOV/2014
 Build 6
 Moved the LCD setup for the EEPROM values of backlighting and contrast to occur after the check for the default
 restore. Minor optimisations in initlcd(), draw_s_meter() (lcd-trx2.c), set_factory_defaults(), get_band(),
 reset_fsense(), get_817_band(). Minor optimisations in serial_test.c A.Ryan - 5B4AIY - 24/JAN/2015

 1.06b
 Build 1
 Considerably simplified the frequency counter calculation, which greatly improved its accuracy. See comments
 in _T3Interrupt() (timers_pwm.c) and get_freq(). The calibration factor adjustment limits give a +/- 10kHz
 correction at 10m. A.Ryan - 5B4AIY - 07/FEB/2015
 Build 2
 Minor optimisation in get_freq() Minor formatting changes in serial_test.c A.Ryan - 5B4AIY - 08/FEB/2015
 Build 3
 Added feature that uses a short push of the PWR button to decrement the main display pages. Modified the main loop
 so that the BAND+/BAND- buttons now auto-repeat. A.Ryan - 5B4AIY - 10/FEB/2015
 Build 4
 Minor optimisation in service.c Fixed minor annoyance. When exiting the service module, the power meter would briefly
 display 0.6W, which would then quickly decay to zero. Moved peak_pwr to a global rather than an embedded static in
 function: disp_fwd_pwr(). Removed calibration and configuration page skip logic. A.Ryan - 5B4AIY - 11/FEB/2015
 Build 5
 Restored the automatic page skip of Serial Test On/Off or Polling Timer settings pages if inconsistent serial port
 mode selection in User Configuration menu. Restored Under/Over Voltage settings page skip in Calibration menu if
 their respective alarms are disabled. A.Ryan - 5B4AIY - 12/FEB/2015
 Build 6
 Another attempt at eliminating the spurious 0.6W display on exiting the service module. A.Ryan - 5B4AIY - 13/FEB/2015

 1.06c
 Build 1
 Added optimisation to the auto-increment and auto-decrement function. The speed at which the UP/DOWN buttons
 now auto-repeat is dependant upon which User Configuration page is selected. In the Calibration & Setup mode
 only page 5 has a slow repeat. A.Ryan - 5B4AIY - 15/FEB/2015
 Build 2
 Removed cfg_rpt[] array and substituted an integer constant. See comment for set_repeat_speed(), and comment
 for defined constant SPEED_ARRAY in juma-pa100.h A.Ryan - 5B4AIY - 16/FEB/2015
 Build 3
 Replaced multiple occurences of button hold/release checks with wait_PWR_BAND_UP_rls() and wait_PWR_rls() which
 saved 39 bytes. Simplified the BAND+/BAND- button logic, saving a further 33 bytes. Simplified the page change
 and auto-repeat logic in service.c. A.Ryan - 5B4AIY - 23/FEB/2015
 Build 4
 Added function second_line(). This eliminates numerous lines of duplicate code. A.Ryan - 5B4AIY - 15/MAR/2015
 Build 5
 Replaced second_line() with more general display_line(), and re-located this function to lcd-trx2.c, replaced
 clear_lcd() with a macro. A.Ryan - 5B4AIY - 19/MAR/2015

 1.06d
 Build 1
 Fixed spurious parameter change that occurs if either the UP or DOWN button is held whilst the fast page select
 function is in operation in either the Calibration or User Configuration modes. A.Ryan - 5B4AIY - 20/MAR/2015
 Build 2
 Added splash screen enable/disable to service menu. Change VERY_LONG_PUSH from 1,500mS to 1,200mS to accord
 with Juma TRX-2. A.Ryan - 5B4AIY - 18/APR/2015
 Build 3
 Added firmware version and build number to System Configuration dump in serial_test.c A.Ryan - 5B4AIY - 19/APR/2015
 Build 4
 Minor optimisation. Removed the direction global and instead passed the parameters to cfg_page_change() and
 change_cfg_page() functions, as well as in the service module. A.Ryan - 5B4AIY - 28/MAY/2015
 Build 5
 Minor re-organisation of boot code. Optimised scaled_temperature(). Fixed a subtle display bug. When switching between
 the Celsius and Fahrenheit scales the graphic limit display was not the same despite the temperature limits being
 equivalent. This is now fixed. Changed the scale factors for the C/F scales. (See notes in juma-pa100.h) Minor
 optimisation of calc_swr() A.Ryan - 5B4AIY - 20/JUN/2015
 Build 6
 Fixed minor bug in service.c which could have allowed a spurious parameter change when switching pages in the fast
 page switch mode. Fixed minor bug in LCD Test in serial_test.c Added a default choice to set_chgen() in lcd-trx.c
 A.Ryan - 5B4AIY - 28/JUN/2015
 Build 7
 Optimised set_chgen() in lcd-trx2.c A.Ryan - 5B4AIY - 01/JUN/2015

 2.00a
 Build 1
 Completely changed the way the checksum is calculated and verified. Added crc_16(), crc_8(), and crc_1() to perform
 a 16-bit cyclic redundancy check. The previous simple addition and complement checksum was not particularly robust,
 and, whilst I have not had any reports of this method failing, it certainly could not detect many types of error.
 The 16-bit CRC on the other hand can detect quite subtle errors. Even so, since it can only generate 65,536 different
 values, it is clearly possible for even this method to fail, but the probability is very low. This version will cause
 a checksum error when first loaded, and the existing calibration and configuration data should be saved prior to loading
 this version. A.Ryan - 5B4AIY - 31/JUL/2015
 Build 2
 Optimised crc_1(). A.Ryan - 5B4AIY - 01/AUG/2015
 Build 3
 Added function get_one_zero() to simplify service and user configuration adjustment code. A.Ryan - 5B4AIY - 05/AUG/2015
 Build 4
 At the suggestion of Olaf, LA3RK, the SWR alarm is now only asserted if the PA State is OPERATE. This avoids nuisance
 SWR alarms in the STANDBY state when tuning an antenna using just the driver transceiver. A.Ryan - 5B4AIY - 09/SEP/2015
 Build 5
 Removed redundant 'old_swr' variable. In the previous firmware it took two main loop cycles for a SWR alarm to be
 registered, but repeated testing has not shown any real need for this, so in this firmware this double sampling has
 been eliminated. This required the maximum sample count to be increased from 8 to at least 16, giving a maximum alarm
 delay of about 80mS. A.Ryan - 5B4AIY - 23/SEP/2015
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "juma-pa100.h"						// JUMA-TRX2 hardware specific definitions
#include "pa100_eeprom.h"					// EEPROM storage structure
#include "DataEEPROM.h"

_FOSC(CSW_FSCM_OFF & XT_PLL4);				// 7,3728MHz XT, 29,4912MHz clock = 7,3728MHz cycle
//_FOSC(CSW_FSCM_OFF & XT_PLL8);			// 3,68640MHz XT, 29,4912MHz clock = 7,3728MHz cycle
_FWDT(WDT_OFF);                				// Turn off the Watch-Dog Timer.
_FBORPOR(MCLR_EN & PWRT_64 & BORV_20);   	// Enable MCLR, power-on timer, brown out 2V
//_FGS(GEN_PROT);            				// Disable Code Protection, old compiler
 _FBS(CODE_PROT_OFF);
 _FSS(CODE_PROT_OFF);
 _FGS(CODE_PROT_OFF);

// External Functions
extern void us_delay(unsigned int);			// Delay routines
extern void ms_delay(unsigned int);

extern void init_timers_pwm(void);			// Timers & PWM DAC
extern void beep(int, int);					// Generate beep, period, length
extern int encoder_get(void);				// Get simulated encoder value
extern void init_adc12(void);
extern int convert_adc12(unsigned int);		// ADC

// LCD functions
extern void initlcd(void);
extern void lcd_cmd(unsigned char);
extern void lcd_putst(register const char *);
extern void display_line(int, const char *);
extern void display_screen(const char *, const char *);
extern void set_ch_bits(char, int);
extern void set_chgen(int);
extern void draw_s_meter(int);
extern void disp_meter(double, int);		// New version - 5B4AIY - 27/JAN/2014

extern void InitUSART1(void);				// Terminal functions
extern void SetUSART1baud(int);				// Baud rate select
extern void ClearUSART1queue(void);			// Clear buffer
extern void putch(char);
extern unsigned char kbhit(void);
extern unsigned char getch(void);
extern unsigned char getche(void);

//extern void init_spi1(void);				// SPI, main board 74HC595 I/O latch (Not used)

extern void service(void);					// Service functions
extern void serial_test(void);				// Serial I/O
extern void dump_eeprom(void);				// Dump EEPROM contents

extern void serial_pa100(void);				// PA100 protocol engine
extern void clear_buffer(void);				// Clear Juma TRX-2 protocol buffer

// Frequency Counter Test
extern unsigned int freq;					// Frequency counter
extern int band_bins[];						// Found samples
extern int freq_sample_ctr;					// Sample counter for F-SENSE mode
extern int freq_ctr;						// Frequency counter averaging counter

// External Data
extern unsigned int button_timer;			// Long push timer, count from set value to zero, global visibility
extern unsigned int band_query_timer;		// Band query timer
extern unsigned int kx3_timeout;			// Elecraft KX-3 Timeout (Defined in timers_pwm.c)
extern unsigned char cmd_buf[];				// Defined in serial_pa100.c
extern int cmd_buf_idx;						// Defined in serial_pa100.c
extern int blink;							// Alarm blink flag
extern int decay_counter;					// Power meter slow decay
extern int poll_resp_rec;					// Poll response (freq set) received from TRX-2/KX3, 1 = received
extern int enc;

// Local Data
void (*rs232_mode)(void);					// Pointer to function taking void and returning void

const char *prompt_msg;

// Baud Rate divisors for 29,4912 MHz osc =  7.3728 MHz FCY 
/*
	Actual baud rate divisors
	Divisor = ((FCY / (Baudrate * 16)) - 1
[0] 1200	383.0
[1] 2400	191.0
[2] 4800	95.0
[3] 9600	47.0
[4] 19200	23.0
[5] 38400	11.0
[6] 57600	7.0
[7] 115200	3.0
*/
const int br_txt[] = {					// Baud Rate messages, indexed by: eeprom.defval.br
							12,			// 1200 Baud
							24,			// 2400 Baud
							48,			// 4800 Baud
							96,			// 9600 Baud
							192,		// 19200 Baud
							384,		// 38400 Baud
							576,		// 57600 Baud
							1152		// 115200 Baud
							};

const char *fan_spd[] = {				// Indexed by: eeprom.defval.Fan_Control
							"Normal",
							"Low",
							"Medium",
							"High"
							};

const char *c_or_f[] = {				// Indexed by: eeprom.defval.Temp_Units
							"Fahrenheit",
							"Celsius"
							};

const char *f_sense[] = {				// Frequency Sense Mode Indexed by: eeprom.defval.bsel_mode
							"Juma-TRX2",// 0
							"KX3",		// 1
							"F-Sense",	// 2
							"FT-817"	// 3
							};

const char *auto_man[] = {				// Indexed by: eeprom.defval.auto_band
							" M ",
							" A "
							};

const char *gain[] = {					// Indexed by eeprom.defval.rf_gain[eeprom.defval.band], and eeprom.defval.band
							"G1",		// 0
							"G2",		// 1
							"G3",		// 2
							"G4"		// 3
							};

const char *band_units[] = {			// Indexed by: eeprom.defval.Band_Units
							"MHz",
							"Metres"
							};

const char *on_off[] = {
							"Off",
							"On"
							};

const char *amp_state[] = {				// Indexed by: pa_state in serial_test.c module.
							"Standby",
							"Operate"
							};

const char *band_select[] = {			// Indexed by: eeprom.defval.auto_band in serial_test.c module.
							"Manual",
							"Auto",
							};

// Band selector display texts
const char *mb_txt[] = {				// Indexed by: eeprom.defval.band
							"  ?!  ",	// Band = 0 	OUT_OF_BAND
							"  160m",	// Band = 1
							"   80m",	// Band = 2
							"   40m",	// Band = 3
							"   30m",	// Band = 4
							"   20m",	// Band = 5
							"   17m",	// Band = 6
							"   15m",	// Band = 7
							"   12m",	// Band = 8
							"   10m",	// Band = 9		MAX_BAND
							"   ?  "	// Band = 10	NOT_KNOWN
							};

const char *bs_txt[] = {				// Indexed by: eeprom.defval.band
							"  ?!  ",	// Band = 0 	OUT_OF_BAND
							"1.8MHz",	// Band = 1
							"3.5MHz",	// Band = 2
							"  7MHz",	// Band = 3
							" 10MHz",	// Band = 4
							" 14MHz",	// Band = 5
							" 18MHz",	// Band = 6
							" 21MHz",	// Band = 7
							" 24MHz",	// Band = 8
							" 28MHz",	// Band = 9		MAX_BAND
							"   ?  "	// Band = 10 	NOT_KNOWN
							};

const char *save_settings_prompt[] = {
							" Save Settings? ",
							" Reset Defaults?"
							};

const char *start_page_select[] =	{			// Indexed by start_page in cfg_14()
							"O/P Power",
							"SWR",
							"Voltage",
							"Current",
							"Temperature"
							};

const char *page_prompt[] = {					// Indexed by: sub_page1 (Line 1)
							"Auto Band Detect",	// 0
							"Serial Speed",		// 1
							"Serial Test",		// 2
							"Polling Timer",	// 3
							"LCD Display",		// 4
							"LCD Display",		// 5
							"SWR Trip",			// 6
							"Temperature",		// 7
							"Fan Control",		// 8
							"Overtemp",			// 9
							"Fan Cut-In Temp",	// 10
							"Band Display",		// 11
							"Graphical Limits",	// 12
							"Graphic Display",	// 13
							"RF Power Meter",	// 14
							"Start-Up Display"	// 15
							};

const char *graph_type[] = {			// Indexed by: eeprom.defval.display_type
							"Original",
							"Large",
							"Small"
							};

const char *poll_cmd[] = {				// Indexed by: eeprom.defval.bsel_mode, as flag passed in check_polling() function
							"=?B\n\r",	// 0 - Juma TRX-2
							"FA;",		// 1 - Elecraft KX-3
							"",			// 2 - Place Holder - Should never be passed this value
							""			// 3 - Place Holder - Should never be passed this value
							};

const char *CS_Msg[] = {				// Indexed by flag in save_settings()
							"   Cancelled!   ",
							"     Saved!     "
						};

const char *pwr_mtr[] = {
							"Watts",
							"dBm"
							};

const char *pwr_mtr_msg[] =	{
							" ---.-W",
							"--.-dBm"
							};

const char temperature_fmt[] = {"%5s:%8i\337%c"};	// Octal 337 = 0xDF = Degree Symbol
const char limit[] = {"Limit"};
const char start[] = {"Start"};
const char units[] = {"Units:%10s"};
const char Yes_No[] = {"PWR:No BAND+:Yes"};
const char firmware[] = {"\n\r%sD Firmware: %s Build: %s Date: %s\n\r"};
const char copyright[] = {"Copyright: Juha Niinikoski - OH2NLT & Matti Hohtola - OH7SV\n\r"};
const char additional_features[] = {"(Additional features and modifications - Adrian Ryan - 5B4AIY)\n\r"};
const char chk_fmt[] = {"\n\rWords   :%5d\n\rChecksum:%5.4X"};

// Standard Display Prompts
const char Hello[] = {"%10s%6s"};
const char SWR_Alarm[] = {" SWR "};
const char CURR_Alarm[] = {" CURR"};
const char TEMP_Alarm[] = {" TEMP"};
const char HI_V_Alarm[] = {" Hi-V"};
const char LO_Batt_Alarm[] = {" Batt"};
const char LO_V_Alarm[] = {" Lo-V"};
const char RS232_Test_Msg[] = {"RS-232 Echo Test"};
const char Awaiting_Data[] = {"Awaiting data..."};
const char Press_PWR_To_Exit[] = {"\n\rRS-232 Test\n\rPress PWR button to exit...\n\r"};
const char RX_Data[] = {"RX Data: %c  0x%2.2X"};
const char RS232_Test_Terminated[] = {"\n\rRS-232 test terminated\n\r"};
const char _2I[] = {"%2i  "};
const char New_Line[] = {"\n\r"};
const char cfg_0_Msg[] = {"Select:%9s"};
const char cfg_1_Msg[] = {"Baud Rate:%4i00"};
const char cfg_2_Msg[] = {"Mode:%11s"};
const char cfg_3_Msg1[] = {"Interval:%3i Sec"};
const char cfg_3_Msg2[] = {"Disabled"};
const char cfg_4_Msg[] = {"Backlight:%6i"};
const char cfg_5_Msg[] = {"Contrast:%7i"};
const char cfg_6_Msg[] = {"Limit:%8.1f:1"};
const char cfg_12_Msg[] = {"Display:%8s"};
const char cfg_13_Msg[] = {"Type:%11s"};
const char cfg_15_Msg[] = {"Page:%11s"};
const char SWR_NA[] = {"SWR--.-"};
const char EEPROM_Chksum[] = {"\n\rEEPROM Checksums\n\rCalibration  : %X\n\rConfiguration: %X\n\rReset Counter: %d"};
const char Chksum_Err[] = {"\n\rEEPROM Checksum Error! Re-loading Factory Defaults"};
const char Chksum_Err_Msg[] = {" Checksum Error "};
const char Loading_Defaults[] = {"Loading Defaults"};
const char Juma_PA100[] = {"JUMA PA100"};
const char OH2NLT_OH7SV[] = {"  OH2NLT OH7SV  "};
const char Calibration_msg[] = {"  Calibration"};
const char Mode_msg[] = {"  Mode%7s"};
const char Display_Next[] = {" DISPLAY = Page "};
const char Oper_Save[] =    {"OPER = Save/Exit"};
const char System_Clk_Msg[] = {"System Clock: %i kHz\n\r"};
const char STBY_Msg[] = {" STBY"};
const char TX_Msg[] = {"  TX "};
const char Err_Msg[] = {" Err "};
const char OPER_Msg[] = {" OPER"};
const char _16s[] = {"%-16s"};
const char KX3_msg1[] = {"KX3 Msg Time-Out"};
const char KX3_msg2[] = {"Set to: 5000 mS "};
const char User_Msg[] = {"      User"};
const char Config_Msg[] = {" Configuration"};
const char freq_fmt[] = {" %6.3f  "};

// Timing
int fcy;						// Clock / kHz
int loop_ctr = 0;				// Alarm blink counter
int dsp_alarm;					// Alarm message display flag (Gets set in display_alarms(), reset in timers_pwm.c)
int beep_flag = FALSE;
int msg_time = 500;				// KX3 message time-out value, mS.
int rep_dly = _FAST;			// UP/DOWN button repeat delay time

// Display Control
int	lcd_mode = NORMAL_DISPLAY_MODE;
int sub_page0 = 0;				// LCD display sub page for lcd_mode0 - Normal Display Start Page
int sub_page1 = 0;				// LCD display sub page for lcd_mode1 - User Configuration Start Page
int max_page = MAX_SUB_PAGE0;
int adjust_flag = TRUE;			// Calibration & Configuration adjustment flag

unsigned char lcdpbuff[20];		// LCD print buffer. Maximum display length is 16 characters, this is a safety margin.

// ADC & Meter Variables
double batt_avg;				// Average battery voltage, in mV
double batt_scale_factor;		// Used in the graphical battery voltage display
double id_cur = 0.0;			// PA Drain current raw value
double vswr = 0.0;				// SWR * 100 (SWR of 2.5:1 = 250)
double pa_temp = 327.6;			// Starting value for 20C/68F. Ensures temperature display quickly settles to the current value.

int scaled_pa_temp;				// Scaled to deg C or deg F depending upon setting.
int batt_raw;					// Raw battery voltage from A-D
int batt_pre_limit;				// Pre-limit warning threshold value

unsigned long pwr_scale_factor;	// Used to calculate maximum power of graphic power meter display.

long input_freq = 0L;			// Used to display the input frequency in the F-SENSE mode.

long fwd_pwr = 0L;				// Relative forward power
long rev_pwr = 0L;				// Relative reverse power
long out_pwr = 0L;				// Scaled output power = (fwd_pwr - rev_pwr) * scale

int sample_counter = 1;			// Power measurement sample counter
int swr = 100;					// Calculated SWR * 100

// PA Status
int pa_state = STANDBY;			// PA Standby/Operate Status, always start in the STANDBY state
unsigned int alarms = 0;		// Alarm bits

int key;						// TX request status - 1 = Tx 0 = Rx
int last_man_band;				// Last selected manual band.
int fan_stop;					// Fan cut-off temperature.
int changed = FALSE;			// Change flag. Set if any user setting is changed.
int temp_units;					// Current temperature units, 0 = Fahrenheit, 1 = Celsius.
int svc_flag = FALSE;			// Flag for Service Mode (Used by service(), and disp_id())
int	fsense_tst = FALSE;			// Flag for F-sense test printouts - Can be Set/Cleared from Serial Test Suite.

unsigned int fd_counter;		// Counts how many times factory defaults have been reloaded.

// Band Select Limits
const unsigned int band_limits[] = {
									LM1,	// 1,500kHz		Band = 0	Invalid if below this frequency
									LM2,	// 2,650kHz		Band = 1	160m
									LM4,	// 5,250kHz		Band = 2	80m
									LM7,	// 8,500kHz		Band = 3	40m
									LM10,	// 12,000kHz	Band = 4	30m
									LM14,	// 16,000kHz	Band = 5	20m
									LM18,	// 19,500kHz	Band = 6	17m
									LM21,	// 22,500kHz	Band = 7	15m
									LM24,	// 26,000kHz	Band = 8	12m
									LM28	// 30,000kHz	Band = 9	10m
									};

// Yaesu 817 band select ADC #
// This table is with 100k || input, 817 internal impedance is about 6,5k
// Band when over this value                   1.8  3.5  7    10   14    18    21    24    28    50    144   430 MHz
//const unsigned int y817_band_limits[13] = {10, 122, 388, 655, 905, 1155, 1423, 1690, 1931, 2172, 2441, 2709, 2962}; 
/*
 This table is with no load (4M7 || input)
 The band is calculated by comparing the ADC value with this table. Starting from 10, if the voltage is greater
 than the entry, the index is the band number. If 10 is returned, then the band is unknown, and the 28MHz filters selected.
 if 0 is returned, then the band is out of range.
 The first column is the actual threshold voltage level.
 The second column is the Yaesu FT-817 band select voltage.
 The voltages in the fourth column represent the measured threshold at which this band was just selected.
*/
const unsigned int y817_band_limits[] = {		// Voltages calculated for a 5.000V reference.
										10,		// 0.012mV	0	Out-Of-Range
										135,	// 0.165mV	1	0.33V 1.8MHz	160m	171mV
										409,	// 0.499mV	2	0.67V 3.5MHz	75/80m	507mV
										683,	// 0.834mV	3	1.00V 7MHz		40m		850mV
										954,	// 1,164V	4	1.33V 10MHz		30m		1.176V
										1228,	// 1.499V	5	1.67V 14MHz		20m		1.513V
										1503,	// 1.835V	6	2.00V 18MHz		17m		1.851V
										1773,	// 2.164V	7	2.33V 21MHz		15m		2.181V
										2047,	// 2.499V	8	2.67V 24MHz		12m		2.522V
										2322,	// 2.835V	9	3.00V 28MHz		10m		2.856V
										2592	// 3.164V	10	Unknown Band			3.186V
										};

const double temp_cal_factor[] = {
								T_MULTF,
								T_MULTC
								};

// User Configuration Data
union
	{
	struct defval defval;
	unsigned int storage[sizeof(struct defval) / 2];
	} eeprom;

// System Calibration Data
union
	{
	struct calval calval;
	unsigned int ee[sizeof(struct calval) / 2];
	} cal;

/*
 CRC16 is a simple 16-bit cyclic redundancy checksum that provides a more robust method of checking data than many other methods.
 The routine returns a 16-bit CRC. Obviously, the largest block of data that can be checked is limited to 65536 bytes. For
 successive calls for larger blocks, the contents of the checksum can be left, and this partial CRC can continue to be used.

 The algorithm used is similar to that in many EPROM programmers and also disc drives. In the following notes, Bit 0 is the
 Least Significant Bit, Bit 7 is the Most Significant Bit.

 The routine calculates the CRC starting with Bit 0, then Bit 7, Bit 6, .... Bit 1.

 It relies on the following steps:

 1.	XOR the current data bit with Bit 1 of the current checksum,
 2.	XOR the result of Step 1 with Bit 3 of the checksum,
 3.	XOR the result of Step 2 with Bit 5 of the checksum,
 4.	XOR the result of Step 3 with Bit 14 of the checksum,
 5.	Multiply the checksum by 2, discarding any carry from the MSB,
 6.	Add the result of Step 4 to the result of Step 5.

 Steps 1 through 6 are repeated for all subsequent bits in the data byte. The entire process is repeated for all subsequent bytes.
 Thus, if the bit values in the CRC register are:

  15  14  13  12  11  10   9   8   7   6   5   4   3   2   1   0
   a   b   c   d   e   f   g   h   i   j   k   l   m   n   o   p

 and x is the current data bit value, y is the partial result, and ^ represents XOR, then the calculation for steps 1 through 4 is:

	y = ( ( ( ( x ^ o ) ^ m ) ^ k ) ^ b )

 After the left shift (multiply by 2) in Step 5, and the addition in step 6, the bits in the CRC register are:

  15  14  13  12  11  10   9   8   7   6   5   4   3   2   1   0
   b   c   d   e   f   g   h   i   j   k   l   m   n   o   p   y

 To give an example, suppose we wish to calculate the CRC for three bytes of data: C3, 55, AA. If we start with the CRC = 0, then
 the following steps will result (The HEX value is the result after the operation):

	DATA     BIT  VALUE	RESULT         CHECKSUM         HEX
	========================================================
	C3        0     1     1      0000 0000 0000 0001    0001
              7     1     1      0000 0000 0000 0011    0003
              6     1     0      0000 0000 0000 0110    0006
              5     0     1      0000 0000 0000 1101    000D
              4     0     1      0000 0000 0001 1011    001B
              3     0     0      0000 0000 0011 0110    0036
              2     0     0      0000 0000 0110 1100    006C
              1     1     1      0000 0000 1101 1001    00D9

	55        0     1     0      0000 0001 1011 0010    01B2
              7     0     0      0000 0011 0110 0100    0364
              6     1     0      0000 0110 1100 1000    06C8
              5     0     1      0000 1101 1001 0001    0D91
              4     1     1      0001 1011 0010 0011    1B23
              3     0     0      0011 0110 0100 0110    3646
              2     1     0      0110 1100 1000 1100    6C8C
              1     0     0      1101 1001 0001 1000    D918

	AA        0     0     0      1011 0010 0011 0000    B230
              7     1     0      0110 0100 0110 0000    6460
              6     0     0      1100 1000 1100 0000    C8C0
              5     1     0      1001 0001 1000 0000    9180
              4     0     0      0010 0011 0000 0000    2300
              3     1     1      0100 0110 0000 0001    4601
              2     0     1      1000 1100 0000 0011    8C03
              1     1     0      0001 1000 0000 0110    1806

 Final Checksum = 0x1806
*/

unsigned int crc_1(unsigned char data, unsigned int csum)
	{
	data &= 0x01;							// Extract data bit
	data ^= (csum >> 1);					// Step 1
	data ^= (csum >> 3);					// Step 2
	data ^= (csum >> 5);					// Step 3
	data ^= (csum >> 14);					// Step 4
	csum <<= 1;								// Step 5
	return csum + (data & 0x0001);			// Step 6
	}

unsigned int crc_8(unsigned char data, unsigned int csum)
	{
	int i = 7;

	csum = crc_1(data, csum);						// Bit 0

	do {
		csum = crc_1((data >> i--), csum);			// Bits 7 through 1
		} while(i);

	return csum;
	}

unsigned int crc_16(unsigned int data, unsigned int csum)
	{
	csum = crc_8((unsigned char)data, csum);		// Low byte
	return crc_8((unsigned char)(data >> 8), csum);	// Low byte plus High byte
	}

// Set factory defaults for EEPROM storage
void set_factory_defaults(void)
	{
	int i = 0;
// User settings
	do	{
		eeprom.defval.rf_gain[i++] = 0;				// Minimum gain
		} while (i < 11);

	eeprom.defval.band = MAX_BAND;					// 28MHz default band
	eeprom.defval.contrast = DEFAULT_CONTRAST;		// 2000
	eeprom.defval.back_light = DEFAULT_BL;			// 300
	eeprom.defval.serial_test = SERIAL_TEST_ON;		// Test Mode On
	eeprom.defval.br = DEFAULT_BAUD_RATE;			// Initial Baud rate = 9600
	eeprom.defval.bsel_mode = FREQ_SENSE;			// 0 = Juma TRX-2, 1 = Elecraft KX3, 2 = Frequency Sense, 3 = FT-817
	eeprom.defval.poll_timer = POLL_TIMER;			// Juma TRX-2/Elecraft KX3 Polling Interval Time, 0 = Off, Default = 5 seconds
	eeprom.defval.swr_limit = SWR_LIMIT;			// SWR alarm trip point
	eeprom.defval.Temp_Units = CELSIUS;				// Temperature units, 1 = C, 0 = F
	eeprom.defval.Fan_Control = NORMAL_SPD;			// Fan Speed Control 0 = Normal, 1 = Low, 2 = Medium, 3 = High
	eeprom.defval.temp_limit = TEMP_LIMITC;			// Temperature alarm
	eeprom.defval.fan_start = FAN_STARTC;			// Fan start point
	eeprom.defval.pa_state = DEFAULT_PWR;			// Not Used, replaced by global flag pa_state. Now used for start-up display page.
	eeprom.defval.auto_band = MANUAL_BAND;			// Band select type, 0 = Manual, 1 = Auto
	eeprom.defval.Band_Units = MHZ;					// 0 = MHz, 1 = Metres
	eeprom.defval.graph_limits = OFF;				// Graphical Limits Display 0 = Off, 1 = On
	eeprom.defval.display_type = ORIGINAL;			// Graphic Display 0 = Original, 1 = Large Scale, 2 = Small Scale
	eeprom.defval.rf_power = _WATT;					// RF Power Meter scaling, 0 = Watts, 1 = dBm

// Calibration values
	cal.calval.id_mult = ID_MULT;					// I = (ADC * ID_MULT) / 200000 
	cal.calval.batt_mult = BATT_MULT;				// V = (ADC * BATT_MULT) / 1000000
	cal.calval.fwd_pwr_mult = FWD_PWR_MULT;			// P = (((ADC * ADC) / 10) * FWD_PWR_MULT) / 10000000
	cal.calval.beep_len = DEFAULT_BEEP_LEN;			// 50mS
	cal.calval.samples = SAMPLE_MIN;				// Default sample count for O/P power and SWR measurements
	cal.calval.splash = ON;							// Display Splash Screen
	cal.calval.alarm_flags = 0x03;					// Under & Over-voltage alarms on
	cal.calval.overvoltage_trip = OVERVOLTAGE_TRIP;	// Nominally 14.5V
	cal.calval.undervoltage_trip = UNDERVOLTAGE_TRIP;// Nominally 11.0V
	cal.calval.max_power = 100;						// Nominal 100W full-scale setting
	cal.calval.freq_cal = FREQ_CAL;					// Nominal frequency meter calibration factor (999985)
	cal.calval.lo_pwr_offset = LO_PWR_OFFSET;		// Low power range correction

// Increment factory default reset counter
	ReadEE(EEPAGE, EE_FD_LOC, &fd_counter, WORD);
	fd_counter++;

// Save factory defaults reset counter
	EraseEE(EEPAGE, EE_FD_LOC, WORD);
	WriteEE(&fd_counter, EEPAGE, EE_FD_LOC, WORD);
	}

// Save defaults
void save_defval(void)
	{
	int i;

	eeprom.defval.d_csum = 0;

	for(i = 0; i < (sizeof(struct defval) / 2); i++)
		{
		if(i < ((sizeof(struct defval) / 2) - 1))
			eeprom.defval.d_csum = crc_16(eeprom.storage[i], eeprom.defval.d_csum);

 		EraseEE(EEPAGE, ((2 * i) + EEDEF), WORD);
		WriteEE(&(eeprom.storage[i]), EEPAGE, ((2 * i) + EEDEF), WORD);			// EEPROM Address 8 high bits, address + physical EEPROM start 16 low bits
		}
// Conditional print only if Serial Test Mode is On, and the serial port is not being used for Auto-Band Detect
 	if((eeprom.defval.serial_test == SERIAL_TEST_ON) && (eeprom.defval.bsel_mode > 1))
		printf(chk_fmt, i, eeprom.defval.d_csum);								// Print EEPROM checksum
	}

// Read defaults
unsigned int read_defval(void)
	{
	int i;
	unsigned int checksum = 0;

	for(i = 0; i < (sizeof(struct defval) / 2); i++)
		{
		ReadEE(EEPAGE, ((2 * i) + EEDEF), &(eeprom.storage[i]), WORD);			// EEPROM Address 8 high bits, address + physical EEPROM start 16 low bits

		if(i < ((sizeof(struct defval) / 2) - 1))
			checksum = crc_16(eeprom.storage[i], checksum);								
		}

// Read factory defaults reset counter
	ReadEE(EEPAGE, EE_FD_LOC, &fd_counter, WORD);

	if(checksum != eeprom.defval.d_csum) return eeprom.defval.d_csum; else return 0;
	}

// Save calibration values
void save_calval(void)
	{
	int i;

	cal.calval.c_csum = 0;

	for(i = 0; i < (sizeof(struct calval) / 2); i++)
		{
		if(i < ((sizeof(struct calval) / 2) - 1))
			cal.calval.c_csum = crc_16(cal.ee[i], cal.calval.c_csum);

		EraseEE(EEPAGE, ((2 * i) + EECAL), WORD);
		WriteEE(&(cal.ee[i]), EEPAGE, ((2 * i) + EECAL), WORD);	// EEPROM Address 8 high bits, address + physical EEPROM start 16 low bits
		}
// Conditional print only if Serial Test Mode is On, and the serial port is not being used for Auto-Band Detect
 	if((eeprom.defval.serial_test == SERIAL_TEST_ON) && (eeprom.defval.bsel_mode > 1))
		printf(chk_fmt, i, cal.calval.c_csum);
	}

// Read calibration values
unsigned int read_calval(void)
	{
	int i;
	unsigned int checksum = 0;

	for(i = 0; i < (sizeof(struct calval) / 2); i++)
		{
		ReadEE(EEPAGE, ((2 * i) + EECAL), &(cal.ee[i]), WORD);	// EEPROM Address 8 high bits, address + physical EEPROM start 16 low bits

		if(i < ((sizeof(struct calval) / 2) - 1))
			checksum = crc_16(cal.ee[i], checksum);
		}

	if(checksum != cal.calval.c_csum) return cal.calval.c_csum; else return 0;
	}

void wait_PWR_BAND_UP_rls(void)
	{
	while(PWR_SW || !BAND_UP);
	}

void wait_PWR_rls(void)
	{
	while(PWR_SW);
	ms_delay(BUTTON_DEBOUNCE);
	}

/*
 This function sets the repeat speed for the UP/DOWN buttons when in the User Configuration mode.
 The integer constant SPEED_ARRAY contains a bit pattern representing the repeat speed. The bit position represents
 the configuration page, and a '1' in that position represents the fast repeat, a '0' represents the slow speed.
 By shifting a 1 left to the bit position represented by the page number contained in sub_page1 and then masking
 the constant SPEED_ARRAY to determine if it contains a '1' or a '0' rep_dly can be assigned either the fast or the
 slow delay constant. A.Ryan - 5B4AIY - 16/FEB/2015
*/
void set_repeat_speed(void)
	{
	rep_dly = (SPEED_ARRAY & (1 << sub_page1)) ? _FAST : _SLOW;
	}

void display_beeps(int page)
	{
	if(page) beep(HZ587_31, cal.calval.beep_len);	// D tone beep for other displays
	else
		{
		beep(HZ466_85, cal.calval.beep_len * 10);	// B-flat tone for basic display
		ms_delay(500);
		}
	}

int int_round(int dividend, int divisor)
	{
	int q, r;

	q = dividend / divisor;
	r = dividend % divisor;

	if(r >= (divisor >> 1)) q++;

	return q;
	}

// Elecraft KX-3 Auto-Band Select Module
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
int get_band(int frequency)
	{
	int i = 0;

	do	{
		if(frequency < band_limits[i])
			break;					// We have found the band,
		} while (++i < NOT_KNOWN);	// otherwise return NOT_KNOWN.

	return i;
	}

void serial_kx3(void)				// Serial Data Handler
	{
	static int used_flag;			// Initialised to zero by default
	char c;
/*
 First check to see if the polling timer timer has timed out. This check is performed in the
 main program in the auto_band_select() function. Then check to see if there are any characters
 in the UART receive buffer with kbhit().

 If there is a character waiting, then we collect it. Under normal circumstances this would
 be the first character, so we check that the buffer index is zero, and the character is the
 lettr F. If it is not, then we simply do nothing and return.

 If it is the letter F, then this is possibly the start of a valid FA data packet, so now we
 insert this character into the buffer, and increment the pointer, and start the message timer.

 When the next character is received the buffer index is now 1, so this letter should be an A.

 If it is not then we clear the buffer and return and await another character.

 If it is, then this is the start of a valid frequency data packet, so we insert this character
 into the buffer, increment the pointer, and continue. We then check to see whether we have
 reached the end of the buffer or if a timeout has occured, in which case we simply dump the
 buffer and restart.

 Assuming we are not at the end of the buffer nor a timeout, eventually we will receive the
 end-of-message character ; and at this point we convert the first buffer entry into a frequency
 and select a band. If the polling timer is non-zero then we set the polled message received flag,
 clear the buffer and start again.
  
 The KX-3 default baud rate is 4800, and as there are 14 bytes to send, then the minimum
 time is: 10 bits/character (8 + start + stop) * 14 * 1/4800 = 29mSec. So setting the
 timeout value to a generous 500mS should be more than enough to ensure reasonable
 operation but still provide a guard against something going wrong.

 At the slowest speed of 1200 baud, the message would take 117mS, so again, there is
 enough margin to ensure that the message can be received, but still protect against
 some comms problem that would cause an incomplete message to be received.

 The reason for not setting the band to NOT_KNOWN if the message ident is wrong is
 that the amplifier may be 'piggy-backed' onto a serial link that is also being used
 by another piece of software, and that there may well be other messages present in
 the data stream. By simply remembering the last valid band setting we avoid constantly
 unsetting the band and forcing the amplifier into standby simply because the message
 is not intended for us. We still have to set the NOT_KNOWN setting in the event of a
 gross error.
 
 If the PA-100D has polling disabled, then to ensure that the amplifier is set to the current
 band of the KX3, a forced FA query command is sent the first time. To ensure that this
 occurs, it is necessary to power the KX3 up first and let it initialise, then power up
 the PA-100D. The forced FA query command will cause the KX3 to return its current
 frequency, and this will set the current band.
*/
	if(used_flag == FALSE && eeprom.defval.poll_timer == FALSE)
		{
		printf(poll_cmd[1]);						// If this is the first time, force a FA query.
		used_flag = TRUE;
		return;
		}

	if(kbhit())										// We've received a character
		{
		c = getch();								// Get character from UART

		if(cmd_buf_idx == 0)						// If we're at the start of the buffer,
			{
			if(c == 'F')							// and if it is a valid start of message,
				{
				cmd_buf[cmd_buf_idx++] = c;			// add the character to the buffer,
				kx3_timeout = msg_time;				// and start the message timer.
				}

			return;									// Otherwise, just return, it's not for us.
			}

		if(cmd_buf_idx == 1 && c != 'A')			// These are not the droids you're looking for,
			{
			cmd_buf_idx = kx3_timeout = 0;			// so reset the index, the message timer,
			cmd_buf[0] = 0;							// and the buffer.
			return;
			}

		cmd_buf[cmd_buf_idx++] = c;					// Character is valid, so add it to the buffer.

		if((cmd_buf_idx > MSG_LEN) || !kx3_timeout)	// Are we about to overrun the buffer or have we timed-out?
			clear_buffer();							// Yes, so, dump the buffer and restart.
		else										// No, so continue.
			{
			if(c == ';')							// We've received the End-Of-Message character of a data packet.
				{
				if(eeprom.defval.poll_timer)
					poll_resp_rec = TRUE;			// If polling is enabled, set the poll response received flag.

				eeprom.defval.band = get_band((int)(atol(cmd_buf + 2) / 1000L));
				kx3_timeout = 0;					// Reset the message timer,
				clear_buffer();						// reset the buffer and index, and restart.
				}	// End IF
			}	// End ELSE
		}	// End IF
	}

/*
 Setting the address of rs232_mode here avoids having to do numerous loop invariant address calculations.
 This somewhat complex logic sets the address of the function to be executed as follows:
 If either the TRX-2 or KX3 frequency sense modes are selected, the function address is set to serial_pa100 for the TRX-2 mode, or serial_kx3
 for the KX-3 mode. If neither of these modes is used, then the address is set to serial_test if the test mode is set to ON by
 eeprom.defval.serial_test, or else it is set to clear the USART queue, effectively discarding any received characters if test mode is off.
*/
void _rs232_mode(void)
	{
	rs232_mode = (eeprom.defval.bsel_mode == JUMA_TRX2)
				? serial_pa100
				: (eeprom.defval.bsel_mode == ELECRAFT_KX3)
				? serial_kx3
				: (eeprom.defval.serial_test)
				? serial_test
				: ClearUSART1queue;
	}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
/* Calculate SWR
 The Voltage Standing Wave Ratio or VSWR can be calculated from the forward and reverse voltages from the directional
 coupler used to sense power. If Vf is the forward voltage, and Vr the reverse voltage, then:
 
	Ro = Vr / Vf
	VSWR = 1 + Ro / 1 - Ro

 In this case, we want the VSWR to be 100 for a VSWR of 1:1 as we are dealing with integer arithmetic, and the final
 value will be scaled by a factor of 100 when displayed.
*/
double calc_swr(long fwd, long rev)
	{
	double Ro;

	if(fwd < PWR_MTR_DEAD_BAND) return 0.0;	// Only calculate if we have some power

	Ro = (double)rev / (double)fwd;
	return ((1.0 + Ro) / (1.0 - Ro)) * 100.0;
	}

// Display Alarms
void display_alarms(void)
	{
	if(beep_flag || !loop_ctr)
		{
		beep(1843, ALARM_BEEP);			// Always sound beep for alarms.
		beep_flag = FALSE;
		}

	if(!loop_ctr)
		{
		loop_ctr = LOOP_COUNT;			// Reset loop counter for next blink cycle
		dsp_alarm = TRUE;				// Display alarm message
		}

	if(dsp_alarm)						// If set, display alarm message
		{
		if(alarms & SWR_AL)
			{
			lcd_putst(SWR_Alarm);
			return;
			}

		if(alarms & CURR_AL)
			{
			lcd_putst(CURR_Alarm);
			return;
			}

		if(alarms & TEMP_AL)
			{
			lcd_putst(TEMP_Alarm);
			return;
			}

		if(alarms & HI_V)
			{
			lcd_putst(HI_V_Alarm);
			return;
			}

		if(alarms & LO_BATT)
			{
			lcd_putst(LO_Batt_Alarm);
			return;
			}

		if(alarms & LO_V)
			{
			lcd_putst(LO_V_Alarm);
			return;
			}
		}
	else								// Otherwise, display blanks
		lcd_spc(5);
	}

// Calculate & display bar graph meter
int scaled_value(void)
	{
	if(eeprom.defval.graph_limits)		// If Graph Limits is set to ON...
		{
		if(sub_page0 == 1)
			{
/*
 SWR Bridge forward voltage is 3.6V @ 100W. This gives an A-D output value of 4096 * (3.6 / 5.0) = 2950
 Forward Power = (A-D * A-D) which at 100W = 2950 * 2950 = 8,702,500. The graphical meter has a full-scale
 of 48, thus the scale factor = 2950 * 2950 * 100 * (Full-Scale Power/100) / 48

  Scale Factor  Full-Scale Power	Voltage		A-D
  ==================================================
	181302			100W				3.6		2950
	199432			110W				3.8		3094
	217432			120W				3.9		3232
	235693			130W				4.1		3364
	253823			140W				4.3		3490
	271953			150W				4.4		3613
*/
			if(swr < 100) return 0;		// This avoids garbage character in graphic display at first position when there is no reasonable output power.
			return int_round((48 * (swr - 99)), (eeprom.defval.swr_limit - 99));
			}
/*
 The display is effectively that of a suppressed-zero voltmeter. The minimum reading is the undervoltage trip limit,
 and the maximum is the over-voltage trip limit. The scale length is therefore: over-voltage - under-voltage
 The fractional length to display is thus: (actual voltage - undervoltage) / scale length
 and the number of bars is: 48 * fractional length
 The battery voltage is measured from a potential divider consisting of a 33K resistor in series with a 10K resistor.
 The fractional factor is therefore: 10 / (33 + 10) = 0.232558
 If we assume an undervoltage trip of 11V and an overvoltage trip of 15V, this gives a scale length of 4V, and as there
 are 8 divisions, this resolves to 0.5V/division. Using the default battery calibration multiplier of 5249, the
 undervoltage trip A-D value is therefore 2095, and the overvoltage A-D value is 2858.
 The scale length is therefore 2858 - 2095 = 763
 If the input voltage for example is 13V, this is an A-D value of 2477, batt_raw would therefore be 2477 * 5250 = 13004250
 Thus the first value is: v = (batt_raw - cal.calval.undervoltage_trip) * cal.calval.batt_mult / BATT_MULT
 The variable value would then be 48 * 2447 / 763 = 117456 / 763 = 153
 Taking account of the scale length, v = v /  (cal.calval.overvoltage_trip - cal.calval.undervoltage_trip)
 Since there are 6 divisions to each scale bar, this gives a fraction display of 153 / 6 = 25, or half scale.
 Note that there is a constant loop-invariant expression here: 48 * cal.calval.batt_mult / BATT_MULT,
 as well as: 48 / (cal.calval.overvoltage_trip - cal.calval.undervoltage_trip)
 It is pointless re-calculating this constant each time through the loop, so it is moved to the initialisation
 section where: batt_scale_factor = cal.calval.overvoltage_trip - cal.calval.undervoltage_trip
*/
		if(sub_page0 == 2) return int_round(48 * (batt_raw - cal.calval.undervoltage_trip), batt_scale_factor);
/*
 The nominal sense voltage for the drain current is 100mV/A.
 id_cur = (double)(convert_adc12(ID_CUR) * (double)cal.calval.id_mult)
 Using the design value of the calibration factor 2442, and for a 24A limit current, the A-D output voltage would be:
 4096 * 2.4 / 5.0 = 1966. The variable: id_cur would therefore be: 1966 * 2442 = 4,800,972.
 The maximum number of bars is 48, thus the scale factor is: 4,800,972 / 48 = 100020.25 i.e, 100020.
 The number of bars to display is therefore id_cur / 100020
*/
		if(sub_page0 == 3) return (int)(id_cur / 100020.0);
		if(sub_page0 == 4)
			{
			if(eeprom.defval.Temp_Units) return int_round((48 * scaled_pa_temp), eeprom.defval.temp_limit);
			else return int_round((48 * (scaled_pa_temp - 32)), (eeprom.defval.temp_limit - 32));
			}

		}
	return (int)(fwd_pwr / pwr_scale_factor);	// Default display, RF O/P Power
	}

void disp_fwd_pwr(void)
	{
	static long peak_pwr;
	double power;

	if(out_pwr >= peak_pwr)						// Fast attack
		{
		peak_pwr = out_pwr;						// Fast attack
		decay_counter = PEAK_SHOW_TIME;			// Reload decay timer (1,000mS)
		}
	else if(decay_counter == 0)
		{
		peak_pwr = 0L;
		power = 0.0;
		}

	power = ((double)peak_pwr * (double)cal.calval.fwd_pwr_mult);	// Scaling, P = ((ADC * ADC) * Calibration Factor) / 100,000,000

	if(eeprom.defval.rf_power)					// Power in dBm = 10 * log10(power in mW)
		{
		if(power > LOW_PWR_LIMIT)				// 26dBm
			disp_meter((log10(power / 100000.0)), DBM);				// The 10 multiplier is embedded in scale_factor[] in lcd-trx2.c
		else
			lcd_putst(pwr_mtr_msg[1]);
		}
	else										// Power in Watts
		{
		if(power > LOW_PWR_LIMIT)				// 0.4W
			disp_meter(power, WATTS);
		else
			lcd_putst(pwr_mtr_msg[0]);
		}
	}

// Measure & display Drain current meter
void disp_id(void)
	{
	static double drain_current;				// Initialised to zero by default.
	static int hold_counter, state;

	if(state != key)							// If we have changed from RX to TX or vice versa,
		{										// then reset the counter and amplifier current data.
		state = key;
		drain_current = 0.0;
		hold_counter = 0;
		}
	else
		{
		id_cur = (double)(convert_adc12(ID_CUR) * (double)cal.calval.id_mult);

		if(id_cur > drain_current)
			{
			drain_current = id_cur;
			hold_counter = 200;
			}
		else
			{
			if(hold_counter) hold_counter--;
			else drain_current = 0.0;
			}

		disp_meter((svc_flag) ? id_cur : drain_current, AMPS);	// I = id_cur / 200000
		}
	}

// Check if any alarms
void check_alarms(void)
	{
	batt_raw = convert_adc12(BATT_CH);
/*
 In the following tests, the battery voltage calibration factor has already been taken into account in the limits,
 and we are comparing corrected A-D convertor values. (See function svc_7() in service.c for an example.)
*/
	if((cal.calval.alarm_flags & 0x01) && (batt_raw > cal.calval.overvoltage_trip))
		alarms |= HI_V;
// If the low-voltage alarm is disabled, then batt_pre_limit is set to 0 as part of the start-up sequence.
	if((batt_raw < batt_pre_limit) && !(alarms & LO_BATT))
		alarms |= LO_BATT;

	if((cal.calval.alarm_flags & 0x02) && (batt_raw < cal.calval.undervoltage_trip))
		alarms |= LO_V;

// Check over current alarm bit & set alarm
	if(OC)
		alarms |= CURR_AL;								// Set alarm flag

// Check PA temperature
	if(scaled_pa_temp > eeprom.defval.temp_limit)
		alarms |= TEMP_AL;								// Set alarm flag

// Check SWR, SWR Alarm only active in the OPERATE state.
	if((swr > eeprom.defval.swr_limit) & pa_state)
		alarms |= SWR_AL;								// Set alarm flag

	if(alarms & 0x0F)									// If SWR, or OC, or High-Temp, or High-Voltage
		{
		TX_ON = OFF;									// RF off immediately
		pa_state = STANDBY;								// Force Standby
		}

	if(alarms) lcd_mode = NORMAL_DISPLAY_MODE;			// Force normal display
	}

// Clear alarms & try to reset over-current latch
void clear_alarms(void)
	{
	loop_ctr = 0;

	if(!alarms) return;

	if(alarms & SWR_AL)
		{
		alarms &= 0x3E;
		return;
		}

	if(alarms & CURR_AL)
		{
		alarms &= 0x3D;							// Clear alarm flag
		OC_CLR = 0;								// Generate reset pulse for over-current latch
		ms_delay(20);
		OC_CLR = 1;
		return;
		}

	if(alarms & TEMP_AL)
		{
		alarms &= 0x3B;
		return;
		}

	if(alarms & HI_V)
		{
		alarms &= 0x37;
		return;
		}

	if(alarms & LO_BATT)
		{
		batt_pre_limit = 0;
		alarms &= 0x2F;
		return;
		}

	if(alarms & LO_V)
		{
		alarms &= 0x1F;
		return;
		}
	}

// Save Current Settings with prompt
/*
 The prompt flag selects either the Save Settings or the Save Defaults message.
 Mode
 1 - Save Calibration Values,
 2 - Save Configuration Values,
 3 - Save both Calibration & Configuration values,
 7 - Restore Factory Defaults and save all
*/
int save_settings(int prompt, int mode)
	{
	int flag = TRUE;

	display_screen(save_settings_prompt[prompt], Yes_No);
	wait_PWR_BAND_UP_rls();					// Wait for button to be released...
	ms_delay(BUTTON_DEBOUNCE);				// Added to prevent spurious cancel command

	while(BAND_UP && !PWR_SW);				// Wait for a BAND+ or PWR button push...

	lcd_cmd(LINE2);

	if(!BAND_UP)							// If BAND+ button was pressed
		{
		if(mode & 4)						// If complete factory default restore
			set_factory_defaults();

		if(mode & 1)
			save_calval();					// Save new System Calibration Settings

		if(mode & 2)						// Save new User Configuration settings
			{
			save_defval();
			changed = FALSE;				// Reset User Settings Changed flag
			}

		beep(HZ698_45, cal.calval.beep_len * 5);
		}
	else
		{
		if(mode & 1) read_calval();			// Restore previous System Calibration Settings

		if(mode & 2) read_defval();			// Restore previous User Configuration Settings

		flag = FALSE;
		beep(HZ587_31, cal.calval.beep_len);
		}

	lcd_putst(CS_Msg[flag]);
	ms_delay(1000);
	wait_PWR_BAND_UP_rls();					// If still pressed, wait for button to be released...
	set_chgen(eeprom.defval.display_type);	// Load bar graph fonts
	return flag;
	}

// Handle PWR button.
void power_off(void)
	{
	button_timer = MEDIUM_PUSH;

	if(alarms)
		{
		clear_alarms();
		wait_PWR_rls();												// This prevents spurious display page change
		}
	else
		{
		do	{
			if(!button_timer)
				{
				beep(HZ392_01, cal.calval.beep_len * 2);			// Play low sound when ready for power down

				if(changed) save_settings(0, 2);					// Prompt to save the current User Settings

				clear_lcd();
				set_pwm3_dac(OFF);									// Turn backlighting off
				FAN1 = OFF;
				FAN2 = OFF;											// Turn fan off
				wait_PWR_rls();										// Wait for power switch to be released,
				PWR_ON = OFF;										// Unlock the power latch,

				while(TRUE);										// and walk off into that long goodnight...
				}
			} while(PWR_SW);

		ms_delay(BUTTON_DEBOUNCE);
		sub_page0--;

		if(sub_page0 < 0) sub_page0 = max_page;

		display_beeps(sub_page0);
		}
	}
 
// RS-232 loop back test (Hold DISPLAY button during power-up)
void rs232_test(void)
	{
	unsigned char c;

	display_screen(RS232_Test_Msg, Awaiting_Data);
	printf(Press_PWR_To_Exit);
	ms_delay(400);							// Ensure message has been sent
	ClearUSART1queue();

	do	{
		if(kbhit())
			{
			c = getch();
			beep(HZ698_45, SHORT_BEEP /5);	// Give beep tone
			putch(c);						// Echo character
			sprintf(lcdpbuff, RX_Data, c, c);
			display_line(LINE2, lcdpbuff);
			}
		} while(!PWR_SW);

	beep(HZ466_85, cal.calval.beep_len);	// Notify PWR button pressed
	printf(RS232_Test_Terminated);
	ClearUSART1queue();
	wait_PWR_rls();							// Wait for switch to be released, this avoids spurious power down
	}

// F-Sense auto band
void reset_fsense(void)						// Clear F-sense 
	{
	int i = 9;

	do	{									// Clear the bins for the next sample set.
		band_bins[i] = 0;
		} while (i--);

	freq_sample_ctr = F_SAMPLES;			// Reload sample counter
	}

void eval_band(int dummy)					// Improved F-sense - A.Ryan - 5B4AIY - 30/APR/2014
	{
	static int last_band;					// Initialised to zero by default
	int i;

	if(!key) last_band = 0;					// Reset last_band in RX mode

	if(!freq_sample_ctr)					// Decide when we have complete set of measurements available
		{
		if(fsense_tst)						// Debug Mode - Can be invoked from Serial Test Suite
			{
			i = 0;

			do	{
				printf(_2I, band_bins[i]);
				} while (++i < 10);

			printf(New_Line);
			}

		i = 9;

		if(band_bins[0] < V_SAMPLES)		// We have meaningful sample set, enough measurements != 0
			{
			do	{
				if(band_bins[i]) break;
				} while (--i);

			if(last_band < i)
				{
				last_band = i;
				eeprom.defval.band = i;		// Set band
				}
			}
		reset_fsense();						// Start next round
		}
	}

int get_817_band(int v)
	{
	int i = 10;

	do	{
		if(v > y817_band_limits[i]) break;	// We have found the band,
		} while (--i);						// otherwise return UNKNOWN.

	return i;
	}
// Check Yaesu 817 band data (voltage)
void eval_y817_band(int dummy)
	{
	eeprom.defval.band = get_817_band(convert_adc12(Y_817));
	}

// Set Gain & Filter Relays
void set_relays(void)
	{
// Set RF Gain relays
/*
 If we write it this way:
	DB_2 = (eeprom.defval.rf_gain[eeprom.defval.band] & 0x01) ? 0 : 1;
	DB_4 = (eeprom.defval.rf_gain[eeprom.defval.band] & 0x02) ? 0 : 1;
 it takes 18 bytes more code! A.Ryan - 5B4AIY - 07/JUN/2013
*/
	if(eeprom.defval.rf_gain[eeprom.defval.band] & 0x01) DB_2 = 0; else DB_2 = 1;
	if(eeprom.defval.rf_gain[eeprom.defval.band] & 0x02) DB_4 = 0; else DB_4 = 1;

// Set RF Filter relays, elegant method
/*
 Now here's an interesting point. If we use this expression to set/clear the relays:
 M1_8 = (eeprom.defval.band == 1);
 or if we use:
 M1_8 = (eeprom.defval.band == 1) ? 1 : 0;
 we get the same code. Interestingly, if we use:
 if(eeprom.defval.band == 1) M1_8 = 1; else M1_8 = 0;
 then the resultant code is 168 bytes shorter! A.Ryan - 5B4AIY - 07/JUN/2013

 The valid band numbers are:

	1	-	160m
	2	-	80m
	3	-	40m
	4	-	30m
	5	-	20m
	6	-	17m
	7	-	15m
	8	-	12m
	9	-	10m

 If the band number is 0, this is out-of-band, and the 10m filter is selected.
 Similarly, if the band number is 10, then the 10m filter is also selected.
*/
	if(eeprom.defval.band == 1)	M1_8 = 1;	else M1_8 = 0;		// Set/Clear 160m relay
	if(eeprom.defval.band == 2) M3_5 = 1;	else M3_5 = 0;		// Set/Clear 80m relay
	if(eeprom.defval.band == 3) M7 = 1;		else M7 = 0;		// Set/Clear 40m relay
	if(eeprom.defval.band == 4) M10 = 1;	else M10 = 0;		// Set/Clear 30m relay
	if((eeprom.defval.band == 5) || (eeprom.defval.band == 6)) M14_18 = 1;	else M14_18 = 0;	// Set/Clear 20m-17m relay
	if((eeprom.defval.band == 0) || (eeprom.defval.band > 6)) M21_28 = 1;	else M21_28 = 0;	// Set/Clear 15m-12m-10m relay
	}

void check_polling(int flag)
	{
	if(eeprom.defval.poll_timer && !band_query_timer)	// Check if polling is enabled and timer has timed-out.
		{
		band_query_timer = eeprom.defval.poll_timer * 1000;	// Yes, so reset the timer, and retry.

		if(poll_resp_rec == FALSE)						// We have not received a valid band set response since
			eeprom.defval.band = NOT_KNOWN;				// the last query, so the band information has been lost.

		printf(poll_cmd[flag]);							// Send the next polling request,
		poll_resp_rec = FALSE;							// and reset the poll response flag.
		}
	}

void (*select_auto_band[])(int) = {						// Indexed by: eeprom.defval.bsel_mode
									check_polling,		// 0 TRX-2
									check_polling,		// 1 KX-3
									eval_band,			// 2 (Dummy Argument)
									eval_y817_band		// 3 (Dummy Argument)
									};

int scaled_temperature(double t)
	{
	t /= temp_cal_factor[eeprom.defval.Temp_Units];

	if(eeprom.defval.Temp_Units == FAHRENHEIT) t += 32.0;

	return (int)(floor(t + 0.5));
	}

// Do analog measurements
void analog_measurements(void)
	{
	static long fwd_avg, rev_avg;

// Measure temperature
	pa_temp = (pa_temp * 0.98) + ((double)convert_adc12(TEMP) * 0.02);	// 50-sample running average
	scaled_pa_temp = scaled_temperature(pa_temp);

// Convert fwd & rev A/D channels
	if(sample_counter--)
		{
		fwd_avg += (long)(convert_adc12(FWD_PWR) + cal.calval.lo_pwr_offset);
		rev_avg += (long)convert_adc12(REV_PWR);
		}
	else
		{
		fwd_avg /= (long)cal.calval.samples;
		rev_avg /= (long)cal.calval.samples;
		fwd_pwr = fwd_avg * fwd_avg;
		rev_pwr = rev_avg * rev_avg;
		out_pwr = fwd_pwr - rev_pwr;
		vswr = calc_swr(fwd_avg, rev_avg);
		swr = (int)vswr;
		fwd_avg = rev_avg = 0L;
		sample_counter = (int)cal.calval.samples;
		}
	}

void max_min(int *value, int max, int min)
	{
	if(*value > max) *value = max;
	if(*value < min) *value = min;
	}
/*
 PA cooling fan control
 This slightly more complex method avoids unnecessary switching of the fan as we progress
 through the various choices. If we simply turned the various control bits on as we climbed
 through the temperature range, on each pass through the main loop we would sequentially switch
 the various fan control bits which might generate a transient. This method ensures that we select
 the desired fan speed immediately, and only switch when necessary. The Fan Control logic now has
 an additional parameter. The user can select whether the fan is on all the time, and the speed
 at which it will run. In the User Configuration menu, the Fan Control option can be set to Normal,
 in which case the fan speed will be controlled entirely by the heat-sink temperature, or Low,
 Medium, or High, in which case the fan will run at the selected speed all the time, and only switch
 to a higher speed as the heatsink temperature rises. Obviously, if you select 'High', then the fan
 will run at this speed all the time. Adrian Ryan - 5B4AIY

 In order to avoid "fan-twitch", the fan speed control is a static variable that retains its value.
 Fan-twitch occured because as the teperature decreased, with no hysteresis, the fan would be commanded
 on and off repeatedly. The fan_stop parameter is calculated once at start-up, or whenever either the
 temperature scale or the start temperature is changed. Adrian Ryan, 5B4AIY
*/
void fan_control(void)
	{
	static unsigned int fan_speed;	// Initialised to zero by default.

	if(scaled_pa_temp <= fan_stop) fan_speed = OFF;

	if(scaled_pa_temp >= eeprom.defval.fan_start) fan_speed = LOW_SPD;
	
	if(eeprom.defval.Temp_Units)										// Celsius Units
		{
		if(scaled_pa_temp >= (eeprom.defval.fan_start + PLUS_5C))		// Start Speed + 5C
			fan_speed = MEDIUM_SPD;

		if(scaled_pa_temp >= (eeprom.defval.fan_start + PLUS_10C))		// Start Speed + 10C
			fan_speed = HIGH_SPD;
		}
	else																// Fahrenheit Units
		{
		if(scaled_pa_temp >= (eeprom.defval.fan_start + PLUS_10F))		// Start Speed + 10F
			fan_speed = MEDIUM_SPD;

		if(scaled_pa_temp >= (eeprom.defval.fan_start + PLUS_20F))		// Start Speed + 20F
			fan_speed = HIGH_SPD;
		}

	fan_speed += eeprom.defval.Fan_Control;								// Select whether fan runs all the time
	max_min(&fan_speed, HIGH_SPD, OFF);
	FAN1 = (fan_speed & 0x02) ? ON : OFF;
	FAN2 = (fan_speed & 0x01) ? ON : OFF;
	}

void display_hdr(void)
	{
	printf(firmware, Juma_PA100, VERSION, BUILD_NUMBER, BUILD_DATE);
	printf(copyright);
	printf(additional_features);
	printf(System_Clk_Msg, fcy);
	}

void set_value(int multiplier, int *value, int max, int min)
	{
	if(adjust_flag) *value += (encoder_get() * multiplier);

	max_min(value, max, min);
	}

void get_one_zero(int *value)
	{
	set_value(1, value, 1, 0);
	}
/*
 The following functions are the user configuration code from the previous version. In all previous versions the actual
 code selected was part of a long switch/case construct that was difficult to read. It was also easy to get lost
 among the various indent levels. One of the attributes of elegant well-written code is the overall feature of burying
 complexity. The higher levels of the code should be simple to read and easy to modify. To this end I decided to remove
 this complex switch/case construct and replace it with a single line of code. This is achieved by placing each page's
 code into a small function, and then collecting these functions into an array, and finally executing the appropriate
 function using the page as an index into the array. Adrian Ryan - 5B4AIY May 2012.
*/
void cfg_0(void)						// Auto band select mode (0 = Juma TRX-2, 1 = Elecraft KX-3, 2 = F-Sense, 3 = FT-817)
	{
	set_value(1, &eeprom.defval.bsel_mode, FT_817, JUMA_TRX2);
	_rs232_mode();						// Set appropriate serial mode
	max_page = (eeprom.defval.bsel_mode == FREQ_SENSE) ? MAX_SUB_PAGE0 + 1 : MAX_SUB_PAGE0;
	sprintf(lcdpbuff, cfg_0_Msg, f_sense[eeprom.defval.bsel_mode]);
	}

void cfg_1(void)						// Set Serial Port Speed (1200 - 115200)
	{
	set_value(1, &eeprom.defval.br, 7, 0);
	SetUSART1baud(eeprom.defval.br);	// Set new Baud rate
	sprintf(lcdpbuff, cfg_1_Msg, br_txt[eeprom.defval.br]);			// Display current selection - Added size of string specifier - 5B4AIY
	}

void cfg_2(void)						// Serial Port Test Mode (Off/On)
	{
	get_one_zero(&eeprom.defval.serial_test);
	_rs232_mode();						// Set appropriate serial mode
	sprintf(lcdpbuff, cfg_2_Msg, on_off[eeprom.defval.serial_test]);
	}

void cfg_3(void)						// Set Polling Interval Time
	{
	set_value(1, &eeprom.defval.poll_timer, MAX_POLL_TIMER, OFF);

	if(eeprom.defval.poll_timer)		// Improved display - 5B4AIY
		sprintf(lcdpbuff, cfg_3_Msg1, eeprom.defval.poll_timer);
	else
		sprintf(lcdpbuff, _16s, cfg_3_Msg2);
	}

void cfg_4(void)						// Set Display Backlighting Level
	{
	set_value(10, &eeprom.defval.back_light, MAX_BL, MIN_BL);
	set_pwm3_dac(eeprom.defval.back_light);
	sprintf(lcdpbuff, cfg_4_Msg,eeprom.defval.back_light);	// Alternative display - 5B4AIY
	}

void cfg_5(void)						// Set Display Contrast Setting
	{
	set_value(50, &eeprom.defval.contrast, MAX_CONTRAST, MIN_CONTRAST);
	set_pwm4_dac(eeprom.defval.contrast);
	sprintf(lcdpbuff, cfg_5_Msg, eeprom.defval.contrast);	// Alternative display - 5B4AIY
	}

void cfg_6(void)						// Set SWR Trip Limit (1.0 - 10.0)
	{
	set_value(10, &eeprom.defval.swr_limit, MAX_SWR, MIN_SWR);
	sprintf(lcdpbuff, cfg_6_Msg, ((double)eeprom.defval.swr_limit) / 100);	// Improved formatting - 5B4AIY
	}

void cfg_7(void)						// Select Display Temperature Units (C/F)
	{
	get_one_zero(&eeprom.defval.Temp_Units);

	if(eeprom.defval.Temp_Units != temp_units)
		{
		temp_units = eeprom.defval.Temp_Units;

		if(eeprom.defval.Temp_Units)
			{
			eeprom.defval.temp_limit = TEMP_LIMITC;
			eeprom.defval.fan_start = FAN_STARTC;
			}
		else
			{
			eeprom.defval.temp_limit = TEMP_LIMITF;
			eeprom.defval.fan_start = FAN_STARTF;						
			}	
		}
	sprintf(lcdpbuff, units, c_or_f[eeprom.defval.Temp_Units]);
	}

void cfg_8(void)						// Select Fan Control Mode (Normal/Low/Medium/High)
	{
	set_value(1, &eeprom.defval.Fan_Control, HIGH_SPD, NORMAL_SPD);
	sprintf(lcdpbuff, cfg_2_Msg, fan_spd[eeprom.defval.Fan_Control]);
	}

void cfg_9(void)						// Set Over Temperature Limit
	{
	char c;

	if(eeprom.defval.Temp_Units)
		{
		set_value(1, &eeprom.defval.temp_limit, MAX_TEMPC, MIN_TEMPC);
		c = 'C';
		}
	else
		{
		set_value(1, &eeprom.defval.temp_limit, MAX_TEMPF, MIN_TEMPF);
		c = 'F';
		}

	sprintf(lcdpbuff, temperature_fmt, limit, eeprom.defval.temp_limit, c);	// Better formatting - 5B4AIY
	}

void cfg_10(void)						// Set Fan Cut-In Temperature
	{
	char c;

	if(eeprom.defval.Temp_Units)		// Celsius
		{
		set_value(1, &eeprom.defval.fan_start, (eeprom.defval.temp_limit - 20), 0);
		fan_stop = eeprom.defval.fan_start - 2;
		c = 'C';
		}
	else								// Fahrenheit
		{
		set_value(1, &eeprom.defval.fan_start, (eeprom.defval.temp_limit - 40), 32);
		fan_stop = eeprom.defval.fan_start - 4;
		c = 'F';
		}

	sprintf(lcdpbuff, temperature_fmt, start, eeprom.defval.fan_start, c);	// Better formatting - 5B4AIY
	}

void cfg_11(void)						// Select Band Units (MHz/Metres)
	{
	get_one_zero(&eeprom.defval.Band_Units);
	sprintf(lcdpbuff, units, band_units[eeprom.defval.Band_Units]);
	}

void cfg_12(void)						// Select graphical display of parameter limits. 
	{
	get_one_zero(&eeprom.defval.graph_limits);
	sprintf(lcdpbuff, cfg_12_Msg, on_off[eeprom.defval.graph_limits]);
	}

void cfg_13(void)						// Select type of graphic display scale
	{
	set_value(1, &eeprom.defval.display_type, SMALL, ORIGINAL);
	sprintf(lcdpbuff, cfg_13_Msg, graph_type[eeprom.defval.display_type]);
	}

void cfg_14(void)						// RF Power Meter Display
	{
	get_one_zero(&eeprom.defval.rf_power);
	sprintf(lcdpbuff, cfg_13_Msg, pwr_mtr[eeprom.defval.rf_power]);
	}

void cfg_15(void)						// Start-Up Page Select
	{
	set_value(1, &start_page, TEMPERATURE, DEFAULT_PWR);
	sprintf(lcdpbuff, cfg_15_Msg, start_page_select[start_page]);
	}

void (*set_cfg[])(void) = {				// Indexed by: sub_page1
						cfg_0,			// 0	Auto Band Select Mode (0 = Juma TRX-2, 1 = Elecraft KX-3, 2 = F-Sense, 3 = FT-817)
						cfg_1,			// 1	Set Serial Port Test Mode (Off/On)
						cfg_2,			// 2	Set Serial Port Speed (1200 - 115200)
						cfg_3,			// 3	Set Polling Interval Time (Off - 10S)
						cfg_4,			// 4	Set Display Backlighting Level (50 - 1000)
						cfg_5,			// 5	Set Display Contrast Setting (0 - 3000)
						cfg_6,			// 6	Set SWR Trip Limit (1.0 - 10.0)
						cfg_7,			// 7	Select Display Temperature Units (C/F)
						cfg_8,			// 8	Select Fan Control Mode (Normal/Low/Medium/High)
						cfg_9,			// 9	Set Over-Temperature Limit
						cfg_10,			// 10	Set Fan Cut-In Temperature
						cfg_11,			// 11	Select Band Units (MHz/Metres)
						cfg_12,			// 12	Select Grphical Limits Display (On/Off)
						cfg_13,			// 13	Select Graphical Display Scale (Original/Large/Small)
						cfg_14,			// 14	Select RF Power Meter Display, (Watts/dBm)
						cfg_15			// 15	Start-Up Page select (Power/SWR/Voltage/Current/Temperature)
						};
/*
 As has been explained in the comments in timers_pwm.c, the constant chosen for the TMR3 interrupt is 7378.
 This gives an actual interrupt timing of 1000.7053uS, but from this must be subtracted the machine cycles
 necessary to stop the counter, transfer the accumulated count, and re-start the counter. Measurements
 indicated that this is about 6 cycles.

 The following table shows the actual frequencies that just result in the tabulated frequency display,
 the frequency correction factor is: 999984

	DISPLAY			ACTUAL FREQUENCY	DIFFERENCE
	==============================================
	1.901				1.900521			+521
	1.899				1.899474			-526

	3.601				3.600524			+524
	3.599				3.599485			-515

	7.101				7.100522			+522
	7.099				7.099484			-516

	10.126				10.125522			+522
	10.124				10.124481			-519

	14.201				14.200524			+524
	14.199				14.199487			-513

	18.119				18.118525			+525
	18.117				18.117469			-531

	21.301				21.300523			+523
	21.299				21.299484			-516

	24.941				24.940527			+527
	24.939				24.939487			-513

	28.851				28.850525			+525
	28.849				28.849485			-515
	==============================================
*/
double get_freq(void)
	{
	static double freq;

	if(!freq_ctr)
		{
		freq = (input_freq > 100000L)	// This eliminates the spurious 0.001 display with no input
			? (double)input_freq / ((double)cal.calval.freq_cal / 10.0)
			: 0.0;
		input_freq = 0;					// Reset frequency averaging counter
		freq_ctr = 100;					// Reset sample counter
		}

	return freq;
	}

void dp_1(void)							// Display SWR
	{
	(swr == 0) ? lcd_putst(SWR_NA) : disp_meter(vswr, SWR);
	}
/*
 The battery voltage is sensed via a potential divider of 33K in series with 10K. The division ratio
 is therefore 10 / 43 = 0.2325581395. The battery calibration factor is scaled to give a battery
 voltage in uV, and for the nominal setting is 5250. Thus, for a nominal amplifier a 13.8V input
 voltage is divided to 13.8 * 0.2325581395 = 3.2093023256, which gives an A-D output of:
 4096 * 3.2093023256 / 5.0 = 2629. The full-scale reading of the voltmeter is 21.5V
 The initial averaging of the battery voltage was 16 samples, but this gave rise to a fair amount
 of jitter at times. After a lot of experimenting the present value of 50 samples was decided. If
 1000 samples were used it would take too long for a true reading to be obtained, 100 samples was
 better, but there was still a significant lag in the readings. 25 samples was not much better than
 16, but 50 samples seems to be just about optimum between reasonable freedom from jitter, and reasonable
 response time. Note that the voltage alarms are virtually instantaneous as they only require a single
 sample. I had given some thought to using the average battery voltage to trip the alarm, but as I'd
 not experienced any problems with transients or other noise triggering alarms, this has not been done
 for this version of the software. A. Ryan 5B4AIY - 30/MAY/2012

 I decided to experiment some more, and I found that the present version using a 50-sample running average
 seems to works the best. It gives a reasonably fast response for large changes, but still smoothes out
 the jitter. A. Ryan - 31/MAY/2012
 Note that the expression for the 50-sample running average: batt_avg = (0.98 * batt_avg) + ((0.02 * voltage) / 1000)
 simplifies to: batt_avg = (0.98 * batt_avg) + (voltage / 50) - A. Ryan - 4/OCT/2012
*/
void dp_2(void)							// Display Supply Voltage
	{
	double voltage;

	voltage = (double)batt_raw * (double)cal.calval.batt_mult;	// Voltage = battery voltage in uV
	batt_avg = (0.98 * batt_avg) + (voltage / 50.0);
	disp_meter(batt_avg, VOLTS);								// batt_avg is in uV
	}

void dp_4(void)							// Display Heat-Sink Temperature
	{
	disp_meter((double)scaled_pa_temp, eeprom.defval.Temp_Units);
	}

void dp_5(void)
	{
	sprintf(lcdpbuff, freq_fmt, get_freq());
	lcd_putst(lcdpbuff);
	disp_fwd_pwr();
	}

void (*display_page[])(void) = {			// Indexed by: sub_page0
							disp_fwd_pwr,	// 0 - Forward Power
							dp_1,			// 1 - SWR
							dp_2,			// 2 - Voltage
							disp_id,		// 3 - Drain Current
							dp_4,			// 4 - Temperature
							dp_5			// 5 - Frequency Display (Extended page, requires F-SENSE mode selected.)
							};

void display_cfg_page(void)
	{
	sprintf(lcdpbuff, _16s, page_prompt[sub_page1]);	// Select page prompt,
	display_line(LINE1, lcdpbuff);			// and display it on line 1.
	set_cfg[sub_page1]();					// Execute selected page code,
	display_line(LINE2, lcdpbuff);			// and display returned value on line 2.
	}

void change_cfg_page(int direction)
	{
	sub_page1 += direction;

	if((eeprom.defval.bsel_mode > 1) && (sub_page1 == 3))		// F-SENSE/FT-817 Mode and Polling Timer page
		sub_page1 += direction;

	if((eeprom.defval.bsel_mode < 2) && (sub_page1 == 2))		// JUMA-TRX2/KX3 Mode and Serial Test On/Off page
		sub_page1 += direction;

	if(sub_page1 > MAX_SUB_PAGE1) sub_page1 = 0;
	if(sub_page1 < 0) sub_page1 = MAX_SUB_PAGE1;

	set_repeat_speed();
	display_beeps(sub_page1);
	}

void cfg_page_change(int direction)
	{
	change_cfg_page(direction);
	adjust_flag = FALSE;		// Prevent spurious adjustments if UP/DOWN button is held
	display_cfg_page();			// whilst fast page repeat is in operation.
	adjust_flag = TRUE;			// Restore normal adjustment.
	ms_delay(PAGE_CHANGE);
	enc = 0;					// Reset encoder.
	}

void range_beep(int value, int upper_limit, int lower_limit)
	{
	if(value == upper_limit) beep(HZ698_45, LONG_BEEP);			// End-of-range F tone beep
	else if(value == lower_limit) beep(HZ466_85, LONG_BEEP);	// End-of-range B-flat tone beep
	else beep(HZ587_31, cal.calval.beep_len);					// Normal D tone beep
	}

// main __________________________________________________________________________________
int main(void)
	{
	int atten;						// Added - 5B4AIY Used to set the input attenuator.
//	int loop_counter = 3125;		// Uncomment to use this to test the loop execution time.
	unsigned int y, w;

// Set general I/O
	TRISA = INIT_TRISA;				// SWITCHES, PTT
	PORTA = INIT_PORTA;

	TRISB = INIT_TRISB;				// ADC inputs, TXEN, RXEN, General I/O
	PORTB = INIT_PORTB;

	TRISC = INIT_TRISC;				// Sideband select & switches
	PORTC = INIT_PORTC;

	TRISD = INIT_TRISD;				// LEDs & LCD pins = output
	PORTD = INIT_PORTD;

	TRISF = INIT_TRISF;				// DDS controls
	PORTF = INIT_PORTF;

	TRISG = INIT_TRISG;				// DDS controls, EEPROM, DScard signals
	PORTG = INIT_PORTG;
// Power switch
	PWR_ON = ON;					// Latch power ON
// Test fan
	FAN2 = ON;						// Test FAN2, low speed on
// Start PWM system
	IPC0 = 0x5444;					// Set tone generator (TMR1) priority higher than others
	init_timers_pwm();				// Setup PWM & tone generators

	fcy = FCY / 1000;				// Clock freq (kHz)

// Init Main Board SPI traffic
//	init_spi1();					// Not used in this application

// Start LCD
	initlcd();						// Initialise LCD with default contrast and backlighting
// Start UART
	InitUSART1();					// Wake up serial interface
// Start ADC
	init_adc12();					// Setup A/D converter, must be done before Keyer is started, RB I/O bits
// Read calibration & default values from EEPROM
	y = read_calval();
	w = read_defval();

 	if((eeprom.defval.serial_test == SERIAL_TEST_ON) && (eeprom.defval.bsel_mode > 1))
		printf(EEPROM_Chksum, y, w, fd_counter);

	if(y || w)						// Compare checksums
		{
		dump_eeprom();				// Test, show EEPROM content before reset
		printf(Chksum_Err);
		display_screen(Chksum_Err_Msg, Loading_Defaults);
		set_factory_defaults();		// EEPROM empty or corrupted, set factory defaults
		save_calval();
		save_defval();
		ms_delay(1500);
		}
// Check if factory defaults asked
	if(!BAND_UP) save_settings(1, 7);	// Check for BAND+ button press
// Set LCD contrast & back light with EEPROM data
	set_pwm4_dac(eeprom.defval.contrast);
	set_pwm3_dac(eeprom.defval.back_light);
	set_chgen(eeprom.defval.display_type);	// Load bar graph fonts
// Check for extension of KX3 message timer for debug purposes
	if(!BAND_DN)
		{
		msg_time = 5000;
		display_screen(KX3_msg1, KX3_msg2);

		while(!BAND_DN);			// Wait for button release
		}
// Put hello messages to LCD if Splash Screen flag is ON
	if((int)cal.calval.splash)
		{
		sprintf(lcdpbuff, Hello, Juma_PA100, VERSION);
		display_screen(lcdpbuff, OH2NLT_OH7SV);
		y = 2000;					// and set sign-on prompt delay time
		}
// Check if service mode start
	button_timer = VERY_LONG_PUSH;	// Set button timer

	while(PWR_SW)					// While PWR button is pressed...
		{
		if(!button_timer)
			{
			beep(HZ392_01, LONG_BEEP);	// Give service beep
			sprintf((char *)lcdpbuff, Mode_msg, VERSION);
			display_screen(Calibration_msg, lcdpbuff);
			wait_PWR_rls();			// Wait for PWR button to be released...
			display_screen(Display_Next, Oper_Save);
			ms_delay(2000);			// and wait for 2 seconds
			svc_flag = TRUE;
			eeprom.defval.band = MAX_BAND;
			service();
			y = 0;					// Set sign-on prompt delay. (No delay required)
			}
		}
// General settings
	if(eeprom.defval.band > MAX_BAND)				// If last shut down was with undefined frequency
		eeprom.defval.band = MAX_BAND;
// Start-up beep
	beep(HZ698_45, SHORT_BEEP * 2);					// Give start-up beep
	FAN1 = ON;										// High speed on, fan test phase 2
	FAN2 = OFF;										// Test FAN2, low speed off
	OC_CLR = 0;										// Clear over-current latch, start pulse
	ms_delay(y);									// Also used as sign-on prompt delay time
	OC_CLR = 1;										// End over-current latch pulse
	FAN1 = OFF;										// Fan test end
	FAN2 = OFF;
// Set user baud rate
	SetUSART1baud(eeprom.defval.br);
// Conditional start up text printout, print only if Serial Test Mode is On
  	if((eeprom.defval.serial_test == SERIAL_TEST_ON) && (eeprom.defval.bsel_mode > 1))
		display_hdr();
// RS-232 I/O Test
	if(!DISP) rs232_test();							// If DISPLAY pressed during startup, Goto RS-232 test loop.
// Other
	temp_units = eeprom.defval.Temp_Units;

	batt_pre_limit = (cal.calval.alarm_flags & 0x02)
		? (int)((long)cal.calval.undervoltage_trip + (100000L / (long)cal.calval.batt_mult))
		: OFF;

	last_man_band = (eeprom.defval.auto_band == MANUAL_BAND)
		? eeprom.defval.band
		: TEN_METRES;

	fan_stop = (eeprom.defval.fan_start - 4) + (2 * eeprom.defval.Temp_Units);				// Hysteresis to stop fan 'twitch', 4F or 2C

	if(eeprom.defval.bsel_mode == FREQ_SENSE)
		max_page++;								// Allow extended Frequency Display page.

	clear_buffer();								// Initialise the Juma TRX-2 receive buffer and index.
	batt_avg = ((double)convert_adc12(BATT_CH) * (double)cal.calval.batt_mult);				// Initial sample for battery voltage, mV
	pwr_scale_factor = (SCALE_CONST * cal.calval.max_power / cal.calval.fwd_pwr_mult);		// See explanation in juma-pa100.h
	batt_scale_factor = cal.calval.overvoltage_trip - cal.calval.undervoltage_trip;
	_rs232_mode();								// Select serial port protocol
	rep_dly = _SLOW;
	decay_counter = 0;							// This seems to help eliminate the spurious 0.6W power
	fwd_pwr = rev_pwr = 0L;						// display when exiting the service mode. A.Ryan 13/FEB/2015
	sub_page0 = start_page;

// Main forever loop _____________________________________________________________________________________________________________
	do	{
/*
 Test counter to determine loop cycle time.
 With the power sampling set to 64, each loop takes 13.1 seconds, thus, 4.19mSec / iteration.
 With the power sampling set to 32, each loop takes 13.0 seconds, thus, 4.16mSec / iteration.
 With the power sampling set to 1, each loop takes 11.6 seconds, thus, 3.56mSec / iteration.
*/
/*
		loop_counter--;

		if(!loop_counter)
			{
			beep(1843, SHORT_BEEP / 5);
			loop_counter = 3125;		// This should give a 10-second loop
			}
*/
/*
 Loop Cycle Time - Measured using an oscilloscope on J19-5, Version 2.0a, Build 5, 23/SEP/2015
 Mode: TRX-2/KX3/F-Sense
 Display
 PWR		4.5mS
 SWR		4.4mS
 Volts		5.3mS
 Amps		5.6mS
 Temp		5.2mS
 Freq/Pwr	4.5mS (Only with F-Sense)
 Mode: FT-817
 PWR		5.0mS
 SWR		4.9mS
 Volts		5.9mS
 Amps		6.2
 Temp		5.8
*/ 
		MAIN_TEST = !MAIN_TEST;			// Timing test J19-5 Check with oscilloscope for reversals
		key = KEY;						// Copy I/O bit to status flag, this uses less code than using the I/O bit directly.
// Auto band select
		select_auto_band[eeprom.defval.bsel_mode](eeprom.defval.bsel_mode);	// Get the current auto band selection,

		if(eeprom.defval.auto_band == MANUAL_BAND)							// and if we are using manual band selection,
			{
			if(eeprom.defval.band == OUT_OF_BAND || eeprom.defval.band == NOT_KNOWN)	// and if it is out-of-band or unknown,
				eeprom.defval.band = last_man_band;							// then restore the last manual band selection.
			}							// Otherwise leave it as it was, i.e, forced auto selection.
// Measurements & basic calculations
		analog_measurements();
// Check alarm conditions
		check_alarms();										// Check if any alarms
		button_timer = LONG_PUSH;							// Set button timer
// Do front panel switches
// DISPLAY / CONFIG
		if(!DISP && !alarms)								// SW1
			{
			do	{
				if(!button_timer)
					{
					if(lcd_mode == NORMAL_DISPLAY_MODE)
						{
						beep(HZ466_85, LONG_BEEP);			// Long beep for major mode change
						lcd_mode = USER_CONFIG_MODE;
						display_screen(User_Msg, Config_Msg);
						set_repeat_speed();
						while(!DISP);						// Wait for button release...
						ms_delay(BUTTON_DEBOUNCE);			// This prevents spurious page change on button release.
						}
					else cfg_page_change(INCREMENT);		// Repeat User Configuration page change whilst button is held...
					}
				} while(!DISP);								// Wait for button release...

			if(button_timer)
				{
				if(lcd_mode == NORMAL_DISPLAY_MODE)			// Select sub page to adjust
					{										// Normal display pages
					sub_page0++;
					sub_page0 %= (max_page + 1);
					display_beeps(sub_page0);
					}
				else change_cfg_page(INCREMENT);			// Single User Configuration page change
				}
			encoder_get();									// Clear encoder after display change
			}

		if(lcd_mode == NORMAL_DISPLAY_MODE)
			{
			if(!alarms)
				{
// OPER
				if(!OPER)										// SW3 - Do not allow change of state in User Config Mode
					{
					pa_state ^= 1;								// Toggle PA State. (Standby = 0, Operate = 1)
					beep((pa_state) ? HZ587_31 : HZ466_85, cal.calval.beep_len);
					while(!OPER);								// Wait for button release...
					ms_delay(BUTTON_DEBOUNCE);
					}
// AUTO
// 11.11.2008 toggle logic
				if(!AUTO)										// SW2
					{
					eeprom.defval.auto_band ^= 1;				// Toggle AUTO/MANUAL Band select. (Manual = 0, Auto = 1)
					changed = TRUE;								// Set configuration changed flag

					if(eeprom.defval.auto_band)					// AUTO
						{
						reset_fsense();							// Reset F-sense logic,
						band_query_timer = 0;					// Force band query if TRX-2 or KX3 protocol selected
						last_man_band = eeprom.defval.band;		// Save current manual band selection,
						eeprom.defval.band = NOT_KNOWN;			// and set band selct to UNKNOWN.
						}
					else eeprom.defval.band = last_man_band;	// MANUAL, so restore last known manual band

					beep((eeprom.defval.auto_band) ? HZ466_85 : HZ587_31, cal.calval.beep_len);
					while(!AUTO);								// Wait for button release...
					ms_delay(BUTTON_DEBOUNCE);
					}
// BAND+ / BAND-
				if((!BAND_UP || !BAND_DN) && !key)				// SW6 - BAND+ SW4 - BAND- Do not allow manual band changes in TX mode!
					{
					changed = TRUE;

					if(eeprom.defval.auto_band)					// AUTO
						{
						eeprom.defval.auto_band = MANUAL_BAND;	// So, change to MANUAL band select mode,
						eeprom.defval.band = last_man_band;		// and restore last known manual band.
						beep(HZ587_31, cal.calval.beep_len);	// B-flat tone beep
						}
					else										// MANUAL
						{
						do	{
							eeprom.defval.band += (!BAND_UP) ? INCREMENT : DECREMENT;
							max_min(&eeprom.defval.band, MAX_BAND, MIN_BAND);
							range_beep(eeprom.defval.band, MAX_BAND, MIN_BAND);
							display_line(LINE2PLUS10, (eeprom.defval.Band_Units) ? mb_txt[eeprom.defval.band] : bs_txt[eeprom.defval.band]);
							ms_delay(PAGE_CHANGE);				// Also serves as the button debounce
							} while(!BAND_UP || !BAND_DN);		// Wait for button release...

						last_man_band = eeprom.defval.band;		// Set the new band.
						}
					}
				}
// Do LCD display
			lcd_cmd(LINE1);							// Start from line 1
// Normal display
// First draw graphical meter if this is not the Extended Frequency Display page
			if(sub_page0 < 5) draw_s_meter(scaled_value());

			display_page[sub_page0]();				// Display selected sub-page
			display_line(LINE2, gain[eeprom.defval.rf_gain[eeprom.defval.band]]);	// Display gain setting
// Show PA state if no alarms, otherwise display alarms
			if(alarms) display_alarms();			// Display alarm message if active
			else
				{
				if(pa_state)						// If there are no alarms, and the PA State is OPERATE,
					{
					if(key)							// and there is a transmit request,
						{
						if((eeprom.defval.band != NOT_KNOWN) && (eeprom.defval.band != OUT_OF_BAND))	// and a valid band is selected,
							prompt_msg = TX_Msg;	// then select the TX message,
						else
							prompt_msg = Err_Msg;	// otherwise select the Error message.
						}
					else prompt_msg = OPER_Msg;						
					}
				else prompt_msg = STBY_Msg;			// If PA State is STANDBY, select the STBY messge.

				lcd_putst(prompt_msg);				// Display the message.
				}
// Show frequency select mode
			lcd_putst(auto_man[eeprom.defval.auto_band]);
// Show selected band
			lcd_putst((eeprom.defval.Band_Units) ? mb_txt[eeprom.defval.band] : bs_txt[eeprom.defval.band]);
// Adjust gain setting, buttons active only in normal display
			atten = encoder_get();

			if(atten)
				{
				changed = TRUE;
				eeprom.defval.rf_gain[eeprom.defval.band] += (atten > 0) ? INCREMENT : (atten < 0) ? DECREMENT : 0;
				max_min(&eeprom.defval.rf_gain[eeprom.defval.band], MAX_GAIN, MIN_GAIN);
				range_beep(eeprom.defval.rf_gain[eeprom.defval.band], MAX_GAIN, MIN_GAIN);
				}
//____________________________________________________________
// Do I/O
// Main board relays
			set_relays();								// Set gain & filter relays
// Evaluate TX possibility
			if(key)										// TX request
				{
				if(pa_state)							// OPER
					{
					if((!alarms) && (eeprom.defval.band != NOT_KNOWN) && (eeprom.defval.band != OUT_OF_BAND))	// No alarms, and valid band selected
						TX_ON = ON;						// RF on
					else
						TX_ON = OFF;					// Something wrong, RF off
					}
				else TX_ON = OFF;						// STANDBY, RF off
				}
			else TX_ON = OFF;							// No TX request, RF off
			}
		else
			{
			pa_state = STANDBY;							// Force state to Standby in User Configuration mode
			display_cfg_page();

			if(!OPER)									// In User Configuration Mode, OPER is the Save & Exit button
				{
				beep(HZ587_31, cal.calval.beep_len);
				save_settings(0, 2);					// Prompt to save User Configuration Settings,
				lcd_mode = NORMAL_DISPLAY_MODE;			// and return to normal display mode.
				rep_dly = _SLOW;						// Slow repeat for Gain changes
				encoder_get();							// Clear encoder after display change
				}
			}
// Fan control
		fan_control();
// Run serial interface functions
		rs232_mode();
// Power switch
		if(PWR_SW)										// SW7
			{
			if(lcd_mode == USER_CONFIG_MODE)			// Increment/Decrement the current User Configuration page
				{
				do	{
					cfg_page_change(DECREMENT);
					} while(PWR_SW);					// Wait for button release...
				}
			else power_off();							//  SW7 - Go to Cancel Alarm/Shutdown/Decrement Display Page
			}
		} while(TRUE);
	}	// End main

