// JUMA-PA100 serial I/O test command set
// Juha Niinikoski, OH2NLT 06.07.2006

#include <stdio.h>					// Added - 5B4AIY
#include <ctype.h>					// Added - 5B4AIY
#include "juma-pa100.h"				// Board definitions
#include "DataEEPROM.h"				// EEPROM I/O definitions, interface to assembler module
#include "pa100_eeprom.h"			// Get EEPROM structure definitions

// External references
// External functions
extern unsigned char kbhit(void);
extern unsigned char getch(void);
extern unsigned long getlong(void);				// see uart.c
extern unsigned char get2hex(void);
extern void ClearUSART1queue();
extern void draw_s_meter(int);					// LCD functions
extern void set_chgen(int);
extern void lcd_cmd(unsigned char);
extern void lcdoutch(unsigned char);			// Added - 5B4AIY
extern void lcd_putst(register const char *);
extern void display_screen(const char *, const char *);
extern int convert_adc12(unsigned int);			// ADC
//extern void exchg_data_spi1(unsigned int);	// Not Used
extern void ms_delay(unsigned int);				// General delay
extern void beep(int, int);						// Added - 5B4AIY
extern int get_817_band(int);					// Get Yaesu 817 band data
extern void display_hdr(void);
extern void save_defaults();
extern void save_calval();
extern unsigned int crc_16(unsigned int, unsigned int);
extern unsigned int crc_8(unsigned char, unsigned int);

// External Data
extern int fsense_tst;				// Flag for F-sense test printouts
extern int fcy;
extern int br_txt[];
extern int alarms;
extern int last_man_band;
extern int pa_state;

extern unsigned char lcdpbuff[];	// LCD print buffer

extern const char firmware[];
extern const char copyright[];
extern const char additional_features[];

extern const char *c_or_f[];
extern const char *f_sense[];
extern const char *band_units[];
extern const char *on_off[];
extern const char *amp_state[];
extern const char *band_select[];
extern const char *bs_txt[];
extern const char *mb_txt[];
extern const char *graph_type[];
extern const char *pwr_mtr[];
extern const char *start_page_select[];

// EEPROM Structures
// User Configuration Data
extern union
	{
	struct defval defval;
	} eeprom;

// System Calibration Data
extern union
	{
	struct calval calval;
	} cal;

// Local Data
const char temp_format[] = {"Sensor:%5.1f%c Displayed:%4.0f%c\n\r"};
const char LCD_Test[] = {"LCD Bar Graph & Character Test"};
const char Write_ASCII[] = {"Write ASCII to LCD, ESC to exit.\n\r"};
const char Write_HEX[] = {"Write HEX to LCD, 1B to exit.\n\r"};
const char ov_msg[] = {"Over-Voltage"};
const char uv_msg[] = {"Under-Voltage"};
const char alarm_is_disabled[] = {" alarm is disabled.\n\r"};
const char amps_volts[] = {"%7.3f %s"};
const char pwr_fmt[] = {"%5.1f Watts (%s)"};
const char temp_fmt[] = {"%5.0f %s"};
const char fan_temp_fmt[] = {"%s%-d%c\n\r"};
const char batt_voltage_dsp[] = {"%4i = %5.2fV\n\r"};
const char band_select_msg[] = {"Last Valid Band Select :%s\n\r"};
const char dividing_line[] = {"----------------------------------------\n\r"};
const char any_key_exits[] = {", any key exits...\n\r"};
const char sensor_cal_msg[] = {"Temperature Sensor Calibration\n\r"};
const char Input_Atten_Status[] = {"Input Attenuator Status\n\r"};
const char Divide_By_Zero[] = {"Divide-By-Zero Trap\n\r"};
const char Test_Terminated[] = {"Test Terminated\n\r"};
const char Alarm_Test[] = {"Alarm System Test\n\r"};
const char fsense_tst_on_off[] = {"\n\rF-Sense Test: %s\n\r"};
const char fsense_hdr[] = {"   160m 80m 40m 30m 20m 17m 15m 12m 10m\n\r"};
const char bad_value[] = {"\n\rIllegal Value!\n\r"};

const char *fan_ctrl[] = {		// Indexed by: eeprom.defval.Fan_Control
						"Normal",
						"Low Speed ON",
						"Medium Speed ON",
						"High Speed ON"
						};

const char *y817[] = {			// Indexed by: j in A-D Convertor test
						"Out-Of-Band!",	// 0
						"1.8MHz/160m",	// 1
						"3.5MHz/80m",	// 2
						"7.0MHz/40m",	// 3
						"10MHz/30m",	// 4
						"14MHz/20m",	// 5
						"18MHz/17m",	// 6
						"21MHz/15m",	// 7
						"24MHz/12m",	// 8
						"28MHz/10m",	// 9
						"Unknown Band!"	// 10
					};

double get_factor(double adc, int cal_factor, double divisor)
	{
	return adc * (double)cal_factor / divisor;
	}

void dump_eeprom(void)
	{
	int i;
	unsigned int w;

	printf("\n\rDump EEPROM contents\n\r%4s%23s\n\r%.23s%.22s", "ADDR", "DATA", dividing_line, dividing_line);

	for(i = 0; i < 256; i += 2)					// Byte address for EEPROM, but data is stored in words
		{
		ReadEE(EEPAGE, (i + EEDEF), &w, WORD);	// EEPROM Address 8 high bits, address + physical EEPROM start 16 low bits

		if((i % 16) == 0) printf("\n\r%4.4X:", i);

		printf("%5.4X", w);			
		}

	printf("\n\r%.23s%.22s\n\r", dividing_line, dividing_line);
	}
/*
// Only Used For Test & Debug
void erase_EEPROM(void)
	{
	int i;

	dump_eeprom();

	for(i = 0; i < 256; i += 2) EraseEE(EEPAGE, (i + EEDEF), WORD);

	dump_eeprom();
	save_defaults();
	save_calval();
	dump_eeprom();
	}
*/

void serial_test(void)
	{
	double a;
	long e, f;
	int i, j, k;
	unsigned char c;

	if(kbhit())
		{
		c = getch();
		switch(c)
			{
// Info commands
			case 'I':
			case '?':
			case 'H':
			case 'h':	// System info
				display_hdr();
				printf("\n\r%.20s[COMMAND TABLE]%.20s\n\r", dividing_line, dividing_line);
				printf("I\tHelp - (This Screen)\n\r");
				printf("A\tADC Channel Dump\n\r");
				printf("a\t%s", Alarm_Test);
				printf("b\t%s\n\r", LCD_Test);
				printf("C\tClear Factory Default Reset Counter\n\r");
//				printf("c\tContinuous SPI write\n\r");		// Not used in this application
				printf("D\tDump System & User Settings\n\r");
				printf("E\tDump EEPROM contents\n\r");
				printf("f\tToggle Frequency Sense Test On/Off\n\r");
				printf("G\t%s", Input_Atten_Status);
				printf("T\t%s", sensor_cal_msg);
				printf("t\tBuzzer Sound Test\n\r");
				printf("W\t%s", Write_ASCII);
				printf("w\t%s", Write_HEX);
				printf("Z\t%s%.15s%s", Divide_By_Zero, dividing_line, dividing_line);
			break;

// Test Commands
			case 'A':	// Measure and display all ADC channels 9 - 14
			printf("\n\r%7s%6s%8s%15s\n\r%.22s%.22s", "CHANNEL", "VALUE", "INPUT", "DISPLAY", dividing_line, dividing_line);
			i = 9;

			do	{
				k = convert_adc12(i);
				printf("\n\rADC%3i:%6i%7.3fV ", i, k, get_factor((double)k, 5, 4096.0));

				switch(i)
					{
					case 9:
						printf(amps_volts, get_factor((double)k, cal.calval.id_mult, 200000.0), "Amps");
					break;

					case 10:
						printf(amps_volts, get_factor((double)k, cal.calval.batt_mult, 1000000.0), "Volts");
					break;

					case 11:
						j = get_817_band(k);
						printf("%3i - %s", j, y817[j]);
					break;

					case 12:
						a = (double)k;
						a *= a;
						printf(pwr_fmt, get_factor(a, cal.calval.fwd_pwr_mult, 100000000.0), "Reverse");
					break;

					case 13:
						a = (double)(k + cal.calval.lo_pwr_offset);
						a *= a;
						printf(pwr_fmt, get_factor(a, cal.calval.fwd_pwr_mult, 100000000.0), "Forward");
					break;

					case 14:
						if(eeprom.defval.Temp_Units)
							printf(temp_fmt, (double)k / T_MULTC, "Celsius");
						else
							printf(temp_fmt, ((double)k / T_MULTF) + 32.0, "Fahrenheit");
					break;
					}
				} while (++i < 15);
				printf("\n\r%.4s%s", dividing_line, dividing_line);
			break;

			case'a':	// Alarm System Test
				printf("\n\r%sEnter 0, 1, 2, 3, 4, 5, or 6 ...\n\r", Alarm_Test);

				c = getch();

				switch(c)
					{
					case '0':
						printf("0 - All Alarms Off\n\r");
						alarms = 0;
					break;

					case '1':
						printf("1 - High SWR\n\r");
						alarms |= SWR_AL;
					break;

					case '2':
						printf("2 - Over-Current\n\r");
						alarms |= CURR_AL;
					break;

					case '3':
						printf("3 - Over-Temperature\n\r");
						alarms |= TEMP_AL;
					break;

					case '4':
						if(cal.calval.alarm_flags & 0x01)
							{
							printf("4 - High Voltage\n\r");
							alarms |= HI_V;
							}
						else
							printf("%s%s", ov_msg, alarm_is_disabled);
					break;

					case '5':
						if(cal.calval.alarm_flags & 0x02)
							{
							printf("5 - Low Battery Pre-Limit\n\r");
							alarms |= LO_BATT;
							}
						else
							printf("%s%s", uv_msg, alarm_is_disabled);
					break;

					case '6':
						if(cal.calval.alarm_flags & 0x02)
							{
							printf("6 - Low Battery Limit\n\r");
							alarms |= LO_V;
							}
						else
							printf("%s%s", uv_msg, alarm_is_disabled);
					break;
					}

				if(alarms & 0x0F)				// Simulate real alarm condition
					{
					TX_ON = OFF;				// RF off immediately
					pa_state = STANDBY;			// Force Standby
					}
			break;

//			case 'B':							// Test & Debug Facility
//				erase_EEPROM();
//			break;

			case 'b':	// Bar Graph Test
				printf("\n\r");
				printf("%s%sSelect Scale: 0,1,2: ", LCD_Test, any_key_exits);
				c = getch() % 3;
				printf("%d - %s\n\r", c, graph_type[c]);
				set_chgen(c);					// Select the desired character generator map, and clear the display.
				c = 0;							// Initialise the character

				while(!kbhit())
					{
					i = 0;
					lcd_cmd(LINE2PLUS4);
					lcdoutch('[');
					lcdoutch(c);
					sprintf((char *)lcdpbuff, " =%3.2X]", c);	// Display LCD characters - Added cast - 5B4AIY
					lcd_putst((char *)lcdpbuff);

					do	{
						lcd_cmd(LINE1PLUS4);
						draw_s_meter(i);
						ms_delay(10);
						} while (++i < 49);

					c++;
					ms_delay(500);
					}

				getch();								// Consume break character
				set_chgen(eeprom.defval.display_type);	// Load the original bar graph fonts
			break;

			case 'C':	// Clear factory default reset counter
				printf("\n\rEEPROM Reset Counter cleared.");
				k = 0;
				EraseEE(EEPAGE, EE_FD_LOC, WORD);
				WriteEE(&k, EEPAGE, EE_FD_LOC, WORD);
				dump_eeprom();
			break;
/*
			case 'c':	// CRC Check (Debug)
				i = crc_8(0xC3, 0x00);
				printf("\n\rData: %.2X  CHECKSUM: %.4X", 0xC3, i);
				i = crc_8(0x55, i);
				printf("\n\rData: %.2X  CHECKSUM: %.4X", 0x55, i);
				i = crc_8(0xAA, i);
				printf("\n\rData: %.2X  CHECKSUM: %.4X", 0xAA, i);
				printf("\n\rFinal Checksum should be: 1806\n\r");
				i = crc_16(0x55C3, 0x00);
				printf("\n\rData: %.4X  CHECKSUM: %.4X", 0x55C3, i);
				i = crc_16(0x00AA, i);
				printf("\n\rData: %.4X  CHECKSUM: %.4X", 0x00AA, i);
				printf("\n\rFinal Checksum should be: 06FA\n\r");
			break;
*/
			case'D':	// Dump System & User Settings
// System Calibration Settings
				printf("\n\r%33s\n\r%7sFirmware: %s Build: %s\n\r%33s", "System Calibration Settings", " ", VERSION, BUILD_NUMBER, dividing_line);
				printf("Battery Voltage Factor : %-4d\n\r", cal.calval.batt_mult);
				printf("PA Current Factor      : %-4d\n\r", cal.calval.id_mult);
				printf("RF Power Meter Factor  : %-4d\n\r", cal.calval.fwd_pwr_mult);
				printf("RF Power Meter Offset  : %-4d\n\r", cal.calval.lo_pwr_offset);
				printf("Beep Length Time       : ");
				(cal.calval.beep_len)
					? printf("%-d mS\n\r", cal.calval.beep_len)
					: printf("OFF\n\r");
				printf("PWR Measurement Samples: ");
				(cal.calval.samples == SAMPLE_MIN) ? printf("1 (Off)\n\r") : printf("%d\n\r", cal.calval.samples);
				printf("Over-Voltage Trip      : ");
				(cal.calval.alarm_flags & 0x01)
					? printf(batt_voltage_dsp, cal.calval.overvoltage_trip, get_factor((double)cal.calval.overvoltage_trip, cal.calval.batt_mult, 1000000.0))
					: printf("%s\n\r", on_off[OFF]);
				printf("Under-Voltage Trip     : ");
				(cal.calval.alarm_flags & 0x02)
					? printf(batt_voltage_dsp, cal.calval.undervoltage_trip, get_factor((double)cal.calval.undervoltage_trip, cal.calval.batt_mult, 1000000.0))
					: printf("%s\n\r", on_off[OFF]);
				printf("Power Meter Full-Scale : %-dW\n\r", cal.calval.max_power);
				printf("Freq Meter Calibration : %-7ld\n\r", cal.calval.freq_cal);
				printf("Splash Screen Display  : %s\n\r", on_off[(int)cal.calval.splash]);
// User Configuration Settings
				printf("\n\r%33s\n\r%s", "User Configuration Settings", dividing_line);
				printf("Frequency Sense Mode   : %s\n\r", f_sense[eeprom.defval.bsel_mode]);
				printf("Serial Link Polling    : ");
				(eeprom.defval.poll_timer)
					? printf("%-d Seconds", eeprom.defval.poll_timer)
					: printf("Off");
				printf("\n\rRS-232 Port Mode       : Test Mode\n\r");
				printf("RS-232 Port Speed      : %-d00 Baud\n\r", br_txt[eeprom.defval.br]);
				printf("LCD Backlighting       : %-4d\n\r", eeprom.defval.back_light);
				printf("LCD Contrast           : %-4d\n\r", eeprom.defval.contrast);
				printf("SWR Trip Limit         : %.2f\n\r", (double)(eeprom.defval.swr_limit) / 100);
				printf("Temperature Units      : %s\n\r", c_or_f[eeprom.defval.Temp_Units]);
				printf("Fan Speed Control      : %s\n\r", fan_ctrl[eeprom.defval.Fan_Control]);
				printf(fan_temp_fmt, "Temperature Alarm Limit: ", eeprom.defval.temp_limit, (eeprom.defval.Temp_Units) ? 'C' : 'F');
				printf(fan_temp_fmt, "Fan Start Temperature  : ", eeprom.defval.fan_start, (eeprom.defval.Temp_Units) ? 'C' : 'F');
				printf("Band Display Units     : %s\n\r", band_units[eeprom.defval.Band_Units]);
				printf("Graphic Limits Display : %s\n\r", on_off[eeprom.defval.graph_limits]);
				printf("Graphic Display Type   : %s\n\r", graph_type[eeprom.defval.display_type]);
				printf("Power Meter Type       : %s\n\r", pwr_mtr[eeprom.defval.rf_power]);
				printf("Start-Up Page          : %s\n\r", start_page_select[start_page]);
				printf("Current Amplifier State: %s\n\r", amp_state[pa_state]);
				printf("Band Select Mode       : %s\n\r", band_select[eeprom.defval.auto_band]);
				printf(band_select_msg, (eeprom.defval.Band_Units) ? mb_txt[last_man_band] : bs_txt[last_man_band]);
				printf("%s", dividing_line);
			break;

			case 'E':	// Dump EEPROM
				dump_eeprom();
			break;

			case 'f':	// Frequency-Sense Test On / Off
				fsense_tst ^= 1;
				if(fsense_tst)
					{
					printf(fsense_tst_on_off, on_off[fsense_tst]);
					printf(fsense_hdr);
					printf("%.39s\n\r", dividing_line);
					}
				else
					{
					printf("%.39s", dividing_line);
					printf(fsense_tst_on_off, on_off[fsense_tst]);
					}
			break;

			case 'G':	// Display Input Attenuator Settings
				printf("\n\r%s", Input_Atten_Status);
				i = 1;

				do	{
					printf("Band: %s G%1i: = %2idB\n\r", bs_txt[i], 1 + eeprom.defval.rf_gain[i], -2 * (3 - eeprom.defval.rf_gain[i]));
					} while (++i < 10);
			break;

			case 'T':	// Calibrate Temperature Sensor
				printf("\n\r%sPress any key to start%s\n\r", sensor_cal_msg, any_key_exits);
				getch();

				while(!kbhit())
					{
					a = i = 0;

					do	{
						a += (double)convert_adc12(TEMP);
						} while (++i < 10);

					if(eeprom.defval.Temp_Units)
						{
						a /= (T_MULTC * 10.0);
						c = 'C';
						}
					else
						{
						a /= (T_MULTF * 10.0);
						a += 32.0;
						c = 'F';
						}

					printf(temp_format, a, c, a, c);
					ms_delay(50);
					}

				getch();						// Flush buffer
				printf(Test_Terminated);
			break;

			case 't':	// Buzzer Sound Test, Pitch = FCY MHz / tone(Hz) * 2
				printf("\n\rSound Test\n\rEnter Frequency (100 - 5000Hz): ");
				e = getlong();

				if(e < 100 || e > 5000)
					{
					printf(bad_value);
					break;
					}

				e = FCY / (e << 1);
				printf("\n\rData Value: %li\n\r", e);
				printf("Enter duration (10 - 32,000mS): ");
				f = getlong();

				if(f < 10 || f > 32000)
					{
					printf(bad_value);
					break;
					}

				beep((int)e, (int)f);
				printf("\n\r");
			break;

			case 'W':	// Test LCD Character Generator, ASCII
				printf("\n\r%s", Write_ASCII);
				clear_lcd();
				i = 0;

				do	{
					c = getch();

					if(i == 16) lcd_cmd(LINE2);
					if(c == 0x0D || i == 32)
						{
						clear_lcd();
						i = 0;
						}

					if(isprint(c))
						{
						printf("Char = 0x%.2X\n\r", c);
						lcdoutch(c);
						i++;
						}
					else printf("Char = 0x%.2X - Non-Printable Character\n\r", c);
					} while(c != 0x1B);

				printf(Test_Terminated);
			break;

			case 'w':	// Test LCD Character Generator, HEX
				printf("\n\r%s", Write_HEX);
				clear_lcd();
				i = 0;

				do	{
					c = get2hex();

					if(i == 16) lcd_cmd(LINE2);
					if(i == 32)
						{
						clear_lcd();
						i = 0;
						}

					lcdoutch(c);

					if(isprint(c))
						{
						i++;
						printf(" = %c\n\r", c);
						}
					else printf(" = Non-Printable Character\n\r");
					} while(c != 0x1B);
					printf(Test_Terminated);
			break;

			case 'Z':	// Trap Divide By Zero Error Test
				printf("\n\r%s", Divide_By_Zero);
				i = 0;
				k /= i;							// Generate error ==> trap
			break;

			default:	// Invalid Command
				printf("\n\rInvalid Command: %c Value: %d\n\r", c, c);
				ClearUSART1queue();				// Flush the receive buffer
			}	// End SWITCH
		}	// End IF
	}	// End FUNCTION

